# File Name: dataImport.py
# Description: Define a class which contains all of the HPA axis data we have
# collected from the literature. The user can create an instance of the class
# for each data set they wish to use in their code
# Author: Christopher Parker
# Created: Wed Aug 17, 2022 | 02:17P EDT
# Last Modified: Mon Sep 12, 2022 | 02:10P EDT

# Importing via __init__.py now
#import numpy as np

class data:
    """
    Imports HPA axis data sets into an object based on which set name is passed.
    Options for set names are:
        - Yehuda (Contains CORT data for MDD, PTSD and healthy control subjects)
        - Carroll (Contains ACTH & CORT data for high-cortisol MDD, low-cortisol
            MDD and healthy control subjects)
        - Golier (Contains ACTH & CORT data for trauma-exposed PTSD,
            trauma-exposed non-PTSD, and healthy control subjects)
        - Bremner (Contains CORT data for abused PTSD, non-abused PTSD, and
            healthy control subjects.)
        - Nelson (Contains ACTH & CORT data for melancholic MDD, atypical MDD,
            uncategorized MDD, and healthy control subjects undergoing
            Trier Social Stress Tests)

    Data sets are imported with time scale in hours by default. To change scale
    to minutes, add "minutes" as the second argument when creating an instance
    of the data class.

    For example:
        nelson = data("nelson", "minutes")
    """
    def __init__(self, dataSet="none", scale="hours"):
        if dataSet.lower() == "yehuda":
            self.yehuda(scale)
        elif dataSet.lower() == "carroll":
            self.carroll(scale)
        elif dataSet.lower() == "golier":
            self.golier(scale)
        elif dataSet.lower() == "bremner":
            self.bremner(scale)
        elif dataSet.lower() == "nelson":
            self.nelson(scale)
        elif dataSet.lower() == "patientf":
            self.patientF(scale)
        else:
            print("Default data set is TSST data in MDD patients from Nelson. To use a different data set, pass the name of the set when creating a new data object.")
            print("For example, data(\"carroll\").")
            print("\nUse help(data) in a Jupyter notebook for documentation including all data set options.")
            self.nelson(scale)


    def __smoothing(self, a, n = 5):
        idx = int((n - 1)/2)
        ret = np.cumsum(a, dtype = float)
        ret[idx + 1:-idx] = ret[n:] - ret[:-n]
        ret[idx] = ret[idx + 2]
        return ret[idx:-idx]/n


    def yehuda(self, scale):
        """
        Contains CORT data for MDD, PTSD and healthy control subjects in the following arrays:
            - controlCortisol
            - depressedCortisol
            - PTSDCortisol

        Each data set can also be retrieved in smoothed form, with each point averaged over the nearest
        5 data points.
        Simply add _smooth to the end of the variable name.

        For example:
            controlCortisol_smooth.
        """
        self.controlCortisol = np.genfromtxt("VeVaPy/data_files/Yehuda-1996-control-cortisol.txt")
        self.PTSDCortisol = np.genfromtxt("VeVaPy/data_files/Yehuda-1996-PTSD-cortisol.txt")
        self.depressedCortisol = np.genfromtxt("VeVaPy/data_files/Yehuda-1996-depressed-cortisol.txt")

        self.controlCortisol_smooth = self.controlCortisol
        self.PTSDCortisol_smooth = self.PTSDCortisol
        self.depressedCortisol_smooth = self.depressedCortisol

        self.controlCortisol_smooth[2:-2,1] = self.__smoothing(self.controlCortisol[:,1])
        self.PTSDCortisol_smooth[2:-2,1] = self.__smoothing(self.PTSDCortisol[:,1])
        self.depressedCortisol_smooth[2:-2,1] = self.__smoothing(self.depressedCortisol[:,1])

        self.controlCortisol = np.genfromtxt("VeVaPy/data_files/Yehuda-1996-control-cortisol.txt")
        self.PTSDCortisol = np.genfromtxt("VeVaPy/data_files/Yehuda-1996-PTSD-cortisol.txt")
        self.depressedCortisol = np.genfromtxt("VeVaPy/data_files/Yehuda-1996-depressed-cortisol.txt")

        # the data time scale is minutes, so convert it to hours when requested
        if scale.lower() == "hours":
            self.controlCortisol[:,0] = np.divide(self.controlCortisol[:,0], 60)
            self.PTSDCortisol[:,0] = np.divide(self.PTSDCortisol[:,0], 60)
            self.depressedCortisol[:,0] = np.divide(self.depressedCortisol[:,0], 60)

            self.controlCortisol_smooth[:,0] = np.divide(self.controlCortisol_smooth[:,0], 60)
            self.PTSDCortisol_smooth[:,0] = np.divide(self.PTSDCortisol_smooth[:,0], 60)
            self.depressedCortisol_smooth[:,0] = np.divide(self.depressedCortisol_smooth[:,0], 60)

    def carroll(self, scale):
        """
        Contains ACTH & CORT data for high-cortisol MDD, low-cortisol MDD, and
        healthy control subjects in the following arrays:
            - controlCortisol, controlACTH
            - HCDepressedCortisol, HCDepressedACTH
            - LCDepressedCortisol, LCDepressedACTH

        Each data set can also be retrieved in smoothed form, with each point
        averaged over the nearest 5 data points.
        Simply add _smooth to the end of the variable name.

        For example:
            controlCortisol_smooth

        Each data set can also be retrieved with the data points rearranged so
        that the data runs from 10AM to 10AM, which matches how the Yehuda data
        is arranged. This is useful for looking at data sets side by side to
        see how the circadian and ultradian rhythms compare.
        Simply add _rearr to the end of the variable name (but before _smooth,
        if present)

        For example:
            controlCortisol_rearr or controlCortisol_rearr_smooth
        """

        self.controlCortisol = np.genfromtxt("VeVaPy/data_files/Carroll-2007-controlGroupCortisol.txt", dtype = float)
        self.HCDepressedCortisol = np.genfromtxt("VeVaPy/data_files/Carroll-2007-HCDepressedCortisol.txt", dtype = float)
        self.LCDepressedCortisol = np.genfromtxt("VeVaPy/data_files/Carroll-2007-LCDepressedCortisol.txt", dtype = float)

        self.controlACTH = np.genfromtxt("VeVaPy/data_files/Carroll-2007-controlGroupACTH.txt", dtype = float)
        self.HCDepressedACTH = np.genfromtxt("VeVaPy/data_files/Carroll-2007-HCDepressedACTH.txt", dtype = float)
        self.LCDepressedACTH = np.genfromtxt("VeVaPy/data_files/Carroll-2007-LCDepressedACTH.txt", dtype = float)

        self.controlCortisol_rearr = np.vstack((self.controlCortisol[60:,:], self.controlCortisol[0:60,:]))
        self.HCDepressedCortisol_rearr = np.vstack((self.HCDepressedCortisol[60:,:], self.HCDepressedCortisol[0:60,:]))
        self.LCDepressedCortisol_rearr = np.vstack((self.LCDepressedCortisol[60:,:], self.LCDepressedCortisol[0:60,:]))

        self.controlACTH_rearr = np.vstack((self.controlACTH[60:,:], self.controlACTH[0:60,:]))
        self.HCDepressedACTH_rearr = np.vstack((self.HCDepressedACTH[60:,:], self.HCDepressedACTH[0:60,:]))
        self.LCDepressedACTH_rearr = np.vstack((self.LCDepressedACTH[60:,:], self.LCDepressedACTH[0:60,:]))

        self.controlCortisol_smooth = self.controlCortisol
        self.HCDepressedCortisol_smooth = self.HCDepressedCortisol
        self.LCDepressedCortisol_smooth = self.LCDepressedCortisol

        self.controlACTH_smooth = self.controlACTH
        self.HCDepressedACTH_smooth = self.HCDepressedACTH
        self.LCDepressedACTH_smooth = self.LCDepressedACTH

        self.controlCortisol_rearr_smooth = self.controlCortisol_rearr
        self.HCDepressedCortisol_rearr_smooth = self.HCDepressedCortisol_rearr
        self.LCDepressedCortisol_rearr_smooth = self.LCDepressedCortisol_rearr

        self.controlACTH_rearr_smooth = self.controlACTH_rearr
        self.HCDepressedACTH_rearr_smooth = self.HCDepressedACTH_rearr
        self.LCDepressedACTH_rearr_smooth = self.LCDepressedACTH_rearr

        self.controlCortisol_smooth[2:-2,1] = self.__smoothing(self.controlCortisol[:,1])
        self.HCDepressedCortisol_smooth[2:-2,1] = self.__smoothing(self.HCDepressedCortisol[:,1])
        self.LCDepressedCortisol_smooth[2:-2,1] = self.__smoothing(self.HCDepressedCortisol[:,1])

        self.controlACTH_smooth[2:-2,1] = self.__smoothing(self.controlACTH[:,1])
        self.HCDepressedACTH_smooth[2:-2,1] = self.__smoothing(self.HCDepressedACTH[:,1])
        self.LCDepressedACTH_smooth[2:-2,1] = self.__smoothing(self.LCDepressedACTH[:,1])

        self.controlCortisol_rearr_smooth[2:-2,1] = self.__smoothing(self.controlCortisol_rearr[:,1])
        self.HCDepressedCortisol_rearr_smooth[2:-2,1] = self.__smoothing(self.HCDepressedCortisol_rearr[:,1])
        self.LCDepressedCortisol_rearr_smooth[2:-2,1] = self.__smoothing(self.LCDepressedCortisol_rearr[:,1])

        self.controlACTH_rearr_smooth[2:-2,1] = self.__smoothing(self.controlACTH_rearr[:,1])
        self.HCDepressedACTH_rearr_smooth[2:-2,1] = self.__smoothing(self.HCDepressedACTH_rearr[:,1])
        self.LCDepressedACTH_rearr_smooth[2:-2,1] = self.__smoothing(self.LCDepressedACTH_rearr[:,1])

        self.controlCortisol = np.genfromtxt("VeVaPy/data_files/Carroll-2007-controlGroupCortisol.txt", dtype = float)
        self.HCDepressedCortisol = np.genfromtxt("VeVaPy/data_files/Carroll-2007-HCDepressedCortisol.txt", dtype = float)
        self.LCDepressedCortisol = np.genfromtxt("VeVaPy/data_files/Carroll-2007-LCDepressedCortisol.txt", dtype = float)

        self.controlACTH = np.genfromtxt("VeVaPy/data_files/Carroll-2007-controlGroupACTH.txt", dtype = float)
        self.HCDepressedACTH = np.genfromtxt("VeVaPy/data_files/Carroll-2007-HCDepressedACTH.txt", dtype = float)
        self.LCDepressedACTH = np.genfromtxt("VeVaPy/data_files/Carroll-2007-LCDepressedACTH.txt", dtype = float)

        self.controlCortisol_rearr = np.vstack((self.controlCortisol[60:,:], self.controlCortisol[0:60,:]))
        self.HCDepressedCortisol_rearr = np.vstack((self.HCDepressedCortisol[60:,:], self.HCDepressedCortisol[0:60,:]))
        self.LCDepressedCortisol_rearr = np.vstack((self.LCDepressedCortisol[60:,:], self.LCDepressedCortisol[0:60,:]))

        self.controlACTH_rearr = np.vstack((self.controlACTH[60:,:], self.controlACTH[0:60,:]))
        self.HCDepressedACTH_rearr = np.vstack((self.HCDepressedACTH[60:,:], self.HCDepressedACTH[0:60,:]))
        self.LCDepressedACTH_rearr = np.vstack((self.LCDepressedACTH[60:,:], self.LCDepressedACTH[0:60,:]))

        # change the time values of the rearranged sets so that 0 minutes is now 10AM
        #  like the non-rearranged sets
        for i in range(len(self.controlCortisol)):
            self.controlCortisol_rearr[i,0] = self.controlCortisol[i,0]
            self.controlACTH_rearr[i,0] = self.controlACTH[i,0]
            self.HCDepressedCortisol_rearr[i,0] = self.HCDepressedCortisol[i,0]
            self.HCDepressedACTH_rearr[i,0] = self.HCDepressedACTH[i,0]
            self.LCDepressedCortisol_rearr[i,0] = self.LCDepressedCortisol[i,0]
            self.LCDepressedACTH_rearr[i,0] = self.LCDepressedACTH[i,0]

            self.controlCortisol_rearr_smooth[i,0] = self.controlCortisol[i,0]
            self.controlACTH_rearr_smooth[i,0] = self.controlACTH[i,0]
            self.HCDepressedCortisol_rearr_smooth[i,0] = self.HCDepressedCortisol[i,0]
            self.HCDepressedACTH_rearr_smooth[i,0] = self.HCDepressedACTH[i,0]
            self.LCDepressedCortisol_rearr_smooth[i,0] = self.LCDepressedCortisol[i,0]
            self.LCDepressedACTH_rearr_smooth[i,0] = self.LCDepressedACTH[i,0]

        if scale.lower() == "hours":
            self.controlCortisol[:,0] = np.divide(self.controlCortisol[:,0], 60)
            self.controlACTH[:,0] = np.divide(self.controlACTH[:,0], 60)
            self.HCDepressedCortisol[:,0] = np.divide(self.HCDepressedCortisol[:,0], 60)
            self.HCDepressedACTH[:,0] = np.divide(self.HCDepressedACTH[:,0], 60)
            self.LCDepressedCortisol[:,0] = np.divide(self.LCDepressedCortisol[:,0], 60)
            self.LCDepressedACTH[:,0] = np.divide(self.LCDepressedACTH[:,0], 60)

            self.controlCortisol_smooth[:,0] = np.divide(self.controlCortisol_smooth[:,0], 60)
            self.controlACTH_smooth[:,0] = np.divide(self.controlACTH_smooth[:,0], 60)
            self.HCDepressedCortisol_smooth[:,0] = np.divide(self.HCDepressedCortisol_smooth[:,0], 60)
            self.HCDepressedACTH_smooth[:,0] = np.divide(self.HCDepressedACTH_smooth[:,0], 60)
            self.LCDepressedCortisol_smooth[:,0] = np.divide(self.LCDepressedCortisol_smooth[:,0], 60)
            self.LCDepressedACTH_smooth[:,0] = np.divide(self.LCDepressedACTH_smooth[:,0], 60)

            self.controlCortisol_rearr[:,0] = np.divide(self.controlCortisol_rearr[:,0], 60)
            self.controlACTH_rearr[:,0] = np.divide(self.controlACTH_rearr[:,0], 60)
            self.HCDepressedCortisol_rearr[:,0] = np.divide(self.HCDepressedCortisol_rearr[:,0], 60)
            self.HCDepressedACTH_rearr[:,0] = np.divide(self.HCDepressedACTH_rearr[:,0], 60)
            self.LCDepressedCortisol_rearr[:,0] = np.divide(self.LCDepressedCortisol_rearr[:,0], 60)
            self.LCDepressedACTH_rearr[:,0] = np.divide(self.LCDepressedACTH_rearr[:,0], 60)

            self.controlCortisol_rearr_smooth[:,0] = np.divide(self.controlCortisol_rearr_smooth[:,0], 60)
            self.controlACTH_rearr_smooth[:,0] = np.divide(self.controlACTH_rearr_smooth[:,0], 60)
            self.HCDepressedCortisol_rearr_smooth[:,0] = np.divide(self.HCDepressedCortisol_rearr_smooth[:,0], 60)
            self.HCDepressedACTH_rearr_smooth[:,0] = np.divide(self.HCDepressedACTH_rearr_smooth[:,0], 60)
            self.LCDepressedCortisol_rearr_smooth[:,0] = np.divide(self.LCDepressedCortisol_rearr_smooth[:,0], 60)
            self.LCDepressedACTH_rearr_smooth[:,0] = np.divide(self.LCDepressedACTH_rearr_smooth[:,0], 60)


    def golier(self, scale):
        """
        Contains ACTH & CORT data for trauma-exposed PTSD, trauma-exposed
        non-PTSD, and healthy control subjects in the following arrays:
            - PTSDCortisol, PTSDACTH
            - nonPTSDTraumaExposedCortisol, nonPTSDTraumaExposedACTH
            - nonPTSDNonExposedCortisol, nonPTSDNonExposedACTH

        Each data set can also be retrieved in smoothed form, with each point
        averaged over the nearest 5 data points.
        Simply add _smooth to the end of the variable name.

        For example:
            PTSDCortisol_smooth.

        Each data set can also be retrieved with the data points rearranged so
        that the data runs from 10AM to 10AM, which matches how the Yehuda data
        is arranged. This is useful for looking at data sets side by side to
        see how the circadian and ultradian rhythms compare.
        Simply add _rearr to the end of the variable name (but before _smooth,
        if present)

        For example:
            PTSDCortisol_rearr or PTSDCortisol_rearr_smooth
        """

        self.PTSDCortisol = np.genfromtxt("VeVaPy/data_files/Golier-2007-PTSD-cortisol.txt", dtype = float)
        self.nonPTSDTraumaExposedCortisol = np.genfromtxt("VeVaPy/data_files/Golier-2007-non-PTSD-trauma-exposed-cortisol.txt", dtype = float)
        self.nonPTSDNonExposedCortisol = np.genfromtxt("VeVaPy/data_files/Golier-2007-non-exposed-control-cortisol.txt", dtype = float)

        self.PTSDACTH = np.genfromtxt("VeVaPy/data_files/Golier-2007-PTSD-ACTH.txt", dtype = float)
        self.nonPTSDTraumaExposedACTH = np.genfromtxt("VeVaPy/data_files/Golier-2007-non-PTSD-trauma-exposed-ACTH.txt", dtype = float)
        self.nonPTSDNonExposedACTH = np.genfromtxt("VeVaPy/data_files/Golier-2007-non-exposed-control-ACTH.txt", dtype = float)

        # rearrange the arrays so that they start at 10AM like the Yehuda data sets
        self.PTSDCortisol_rearr = np.vstack((self.PTSDCortisol[7:,:], self.PTSDCortisol[0:7,:]))
        self.nonPTSDTraumaExposedCortisol_rearr = np.vstack((self.nonPTSDTraumaExposedCortisol[7:,:], self.nonPTSDTraumaExposedCortisol[0:7,:]))
        self.nonPTSDNonExposedCortisol_rearr = np.vstack((self.nonPTSDNonExposedCortisol[7:,:], self.nonPTSDNonExposedCortisol[0:7,:]))

        self.PTSDACTH_rearr = np.vstack((self.PTSDACTH[3:,:], self.PTSDACTH[0:3,:]))
        self.nonPTSDTraumaExposedACTH_rearr = np.vstack((self.nonPTSDTraumaExposedACTH[3:,:], self.nonPTSDTraumaExposedACTH[0:3,:]))
        self.nonPTSDNonExposedACTH_rearr = np.vstack((self.nonPTSDNonExposedACTH[3:,:], self.nonPTSDNonExposedACTH[0:3,:]))

        # create the smoothed arrays
        self.PTSDCortisol_smooth = self.PTSDCortisol
        self.nonPTSDTraumaExposedCortisol_smooth = self.nonPTSDTraumaExposedCortisol
        self.nonPTSDNonExposedCortisol_smooth = self.nonPTSDNonExposedCortisol
        self.PTSDACTH_smooth = self.PTSDACTH
        self.nonPTSDTraumaExposedACTH_smooth = self.nonPTSDTraumaExposedACTH
        self.nonPTSDNonExposedACTH_smooth = self.nonPTSDNonExposedACTH

        self.PTSDCortisol_smooth[2:-2,1] = self.__smoothing(self.PTSDCortisol[:,1])
        self.nonPTSDTraumaExposedCortisol_smooth[2:-2,1] = self.__smoothing(self.nonPTSDTraumaExposedCortisol[:,1])
        self.nonPTSDNonExposedCortisol_smooth[2:-2,1] = self.__smoothing(self.nonPTSDNonExposedCortisol[:,1])

        self.PTSDACTH_smooth[2:-2,1] = self.__smoothing(self.PTSDACTH[:,1])
        self.nonPTSDTraumaExposedACTH_smooth[2:-2,1] = self.__smoothing(self.nonPTSDTraumaExposedACTH[:,1])
        self.nonPTSDNonExposedACTH_smooth[2:-2,1] = self.__smoothing(self.nonPTSDNonExposedACTH[:,1])

        self.PTSDCortisol_rearr_smooth = self.PTSDCortisol_rearr
        self.nonPTSDTraumaExposedCortisol_rearr_smooth = self.nonPTSDTraumaExposedCortisol_rearr
        self.nonPTSDNonExposedCortisol_rearr_smooth = self.nonPTSDNonExposedCortisol_rearr
        self.PTSDACTH_rearr_smooth = self.PTSDACTH_rearr
        self.nonPTSDTraumaExposedACTH_rearr_smooth = self.nonPTSDTraumaExposedACTH_rearr
        self.nonPTSDNonExposedACTH_rearr_smooth = self.nonPTSDNonExposedACTH_rearr

        self.PTSDCortisol_rearr_smooth[2:-2,1] = self.__smoothing(self.PTSDCortisol_rearr[:,1])
        self.nonPTSDTraumaExposedCortisol_rearr_smooth[2:-2,1] = self.__smoothing(self.nonPTSDTraumaExposedCortisol_rearr[:,1])
        self.nonPTSDNonExposedCortisol_rearr_smooth[2:-2,1] = self.__smoothing(self.nonPTSDNonExposedCortisol_rearr[:,1])

        self.PTSDACTH_rearr_smooth[2:-2,1] = self.__smoothing(self.PTSDACTH_rearr[:,1])
        self.nonPTSDTraumaExposedACTH_rearr_smooth[2:-2,1] = self.__smoothing(self.nonPTSDTraumaExposedACTH_rearr[:,1])
        self.nonPTSDNonExposedACTH_rearr_smooth[2:-2,1] = self.__smoothing(self.nonPTSDNonExposedACTH_rearr[:,1])

        # re-run the genfromtxt() commands because the smoothing overwrites the non-smoothed
        #  arrays also, for some reason
        self.PTSDCortisol = np.genfromtxt("VeVaPy/data_files/Golier-2007-PTSD-cortisol.txt", dtype = float)
        self.nonPTSDTraumaExposedCortisol = np.genfromtxt("VeVaPy/data_files/Golier-2007-non-PTSD-trauma-exposed-cortisol.txt", dtype = float)
        self.nonPTSDNonExposedCortisol = np.genfromtxt("VeVaPy/data_files/Golier-2007-non-exposed-control-cortisol.txt", dtype = float)

        self.PTSDACTH = np.genfromtxt("VeVaPy/data_files/Golier-2007-PTSD-ACTH.txt", dtype = float)
        self.nonPTSDTraumaExposedACTH = np.genfromtxt("VeVaPy/data_files/Golier-2007-non-PTSD-trauma-exposed-ACTH.txt", dtype = float)
        self.nonPTSDNonExposedACTH = np.genfromtxt("VeVaPy/data_files/Golier-2007-non-exposed-control-ACTH.txt", dtype = float)

        # rearrange the arrays so that they start at 10AM like the Yehuda data sets
        self.PTSDCortisol_rearr = np.vstack((self.PTSDCortisol[7:,:], self.PTSDCortisol[0:7,:]))
        self.nonPTSDTraumaExposedCortisol_rearr = np.vstack((self.nonPTSDTraumaExposedCortisol[7:,:], self.nonPTSDTraumaExposedCortisol[0:7,:]))
        self.nonPTSDNonExposedCortisol_rearr = np.vstack((self.nonPTSDNonExposedCortisol[7:,:], self.nonPTSDNonExposedCortisol[0:7,:]))

        self.PTSDACTH_rearr = np.vstack((self.PTSDACTH[3:,:], self.PTSDACTH[0:3,:]))
        self.nonPTSDTraumaExposedACTH_rearr = np.vstack((self.nonPTSDTraumaExposedACTH[3:,:], self.nonPTSDTraumaExposedACTH[0:3,:]))
        self.nonPTSDNonExposedACTH_rearr = np.vstack((self.nonPTSDNonExposedACTH[3:,:], self.nonPTSDNonExposedACTH[0:3,:]))

        # change the time values of the rearranged sets so that 0 minutes is now 10AM
        #  like the non-rearranged sets
        for i in range(len(self.PTSDCortisol[:,0])):
            self.PTSDCortisol_rearr[i,0] = self.PTSDCortisol[i,0]
            self.PTSDCortisol_rearr_smooth[i,0] = self.PTSDCortisol[i,0]
            self.nonPTSDTraumaExposedCortisol_rearr[i,0] = self.nonPTSDTraumaExposedCortisol[i,0]
            self.nonPTSDTraumaExposedCortisol_rearr_smooth[i,0] = self.nonPTSDTraumaExposedCortisol[i,0]
            self.nonPTSDNonExposedCortisol_rearr[i,0] = self.nonPTSDNonExposedCortisol[i,0]
            self.nonPTSDNonExposedCortisol_rearr_smooth[i,0] = self.nonPTSDNonExposedCortisol[i,0]

        for i in range(len(self.PTSDACTH[:,0])):
            self.PTSDACTH_rearr[i,0] = self.PTSDACTH[i,0]
            self.PTSDACTH_rearr_smooth[i,0] = self.PTSDACTH[i,0]
            self.nonPTSDTraumaExposedACTH_rearr[i,0] = self.nonPTSDTraumaExposedACTH[i,0]
            self.nonPTSDTraumaExposedACTH_rearr_smooth[i,0] = self.nonPTSDTraumaExposedACTH[i,0]
            self.nonPTSDNonExposedACTH_rearr[i,0] = self.nonPTSDNonExposedACTH[i,0]
            self.nonPTSDNonExposedACTH_rearr_smooth[i,0] = self.nonPTSDNonExposedACTH[i,0]

        if scale.lower() == "minutes":
            self.PTSDCortisol[:,0] = np.multiply(self.PTSDCortisol[:,0], 60)
            self.PTSDACTH[:,0] = np.multiply(self.PTSDACTH[:,0], 60)
            self.nonPTSDTraumaExposedCortisol[:,0] = np.multiply(self.nonPTSDTraumaExposedCortisol[:,0], 60)
            self.nonPTSDTraumaExposedACTH[:,0] = np.multiply(self.nonPTSDTraumaExposedACTH[:,0], 60)
            self.nonPTSDNonExposedCortisol[:,0] = np.multiply(self.nonPTSDNonExposedCortisol[:,0], 60)
            self.nonPTSDNonExposedACTH[:,0] = np.multiply(self.nonPTSDNonExposedACTH[:,0], 60)

            self.PTSDCortisol_smooth[:,0] = np.multiply(self.PTSDCortisol_smooth[:,0], 60)
            self.PTSDACTH_smooth[:,0] = np.multiply(self.PTSDACTH_smooth[:,0], 60)
            self.nonPTSDTraumaExposedCortisol_smooth[:,0] = np.multiply(self.nonPTSDTraumaExposedCortisol_smooth[:,0], 60)
            self.nonPTSDTraumaExposedACTH_smooth[:,0] = np.multiply(self.nonPTSDTraumaExposedACTH_smooth[:,0], 60)
            self.nonPTSDNonExposedCortisol_smooth[:,0] = np.multiply(self.nonPTSDNonExposedCortisol_smooth[:,0], 60)
            self.nonPTSDNonExposedACTH_smooth[:,0] = np.multiply(self.nonPTSDNonExposedACTH_smooth[:,0], 60)

            self.PTSDCortisol_rearr[:,0] = np.multiply(self.PTSDCortisol_rearr[:,0], 60)
            self.PTSDACTH_rearr[:,0] = np.multiply(self.PTSDACTH_rearr[:,0], 60)
            self.nonPTSDTraumaExposedCortisol_rearr[:,0] = np.multiply(self.nonPTSDTraumaExposedCortisol_rearr[:,0], 60)
            self.nonPTSDTraumaExposedACTH_rearr[:,0] = np.multiply(self.nonPTSDTraumaExposedACTH_rearr[:,0], 60)
            self.nonPTSDNonExposedCortisol_rearr[:,0] = np.multiply(self.nonPTSDNonExposedCortisol_rearr[:,0], 60)
            self.nonPTSDNonExposedACTH_rearr[:,0] = np.multiply(self.nonPTSDNonExposedACTH_rearr[:,0], 60)

            self.PTSDCortisol_rearr_smooth[:,0] = np.multiply(self.PTSDCortisol_rearr_smooth[:,0], 60)
            self.PTSDACTH_rearr_smooth[:,0] = np.multiply(self.PTSDACTH_rearr_smooth[:,0], 60)
            self.nonPTSDTraumaExposedCortisol_rearr_smooth[:,0] = np.multiply(self.nonPTSDTraumaExposedCortisol_rearr_smooth[:,0], 60)
            self.nonPTSDTraumaExposedACTH_rearr_smooth[:,0] = np.multiply(self.nonPTSDTraumaExposedACTH_rearr_smooth[:,0], 60)
            self.nonPTSDNonExposedCortisol_rearr_smooth[:,0] = np.multiply(self.nonPTSDNonExposedCortisol_rearr_smooth[:,0], 60)
            self.nonPTSDNonExposedACTH_rearr_smooth[:,0] = np.multiply(self.nonPTSDNonExposedACTH_rearr_smooth[:,0], 60)


    def bremner(self, scale):
        """
        Contains CORT data for high-cortisol MDD, low-cortisol MDD, and healthy
        control subjects in the following arrays:
            - abusedPTSDCortisol
            - nonAbusedPTSDCortisol
            - nonAbusedNonPTSDCortisol

        Each data set can also be retrieved in smoothed form, with each point
        averaged over the nearest 5 data points.
        Simply add _smooth to the end of the variable name.

        For example:
            abusedPTSDCortisol_smooth

        Each data set can also be retrieved with the data points rearranged so
        that the data runs from 10AM to 10AM, which matches how the Yehuda data
        is arranged. This is useful for looking at data sets side by side to
        see how the circadian and ultradian rhythms compare.
        Simply add _rearr to the end of the variable name (but before _smooth,
        if present)

        For example:
            abusedPTSDCortisol_rearr or abusedPTSDCortisol_rearr_smooth
        """

        self.abusedPTSDCortisol = np.genfromtxt("VeVaPy/data_files/Bremner-2007-abused-PTSD-cortisol.txt", dtype = float)
        self.nonAbusedPTSDCortisol = np.genfromtxt("VeVaPy/data_files/Bremner-2007-non-abused-PTSD-cortisol.txt", dtype = float)
        self.nonAbusedNonPTSDCortisol = np.genfromtxt("VeVaPy/data_files/Bremner-2007-non-abused-non-PTSD-cortisol.txt", dtype = float)

        # rearrange the data so that we start at 10AM like the Yehuda data
        self.abusedPTSDCortisol_rearr = np.vstack((self.abusedPTSDCortisol[68:,:], self.abusedPTSDCortisol[0:68,:]))
        self.nonAbusedPTSDCortisol_rearr = np.vstack((self.nonAbusedPTSDCortisol[68:,:], self.nonAbusedPTSDCortisol[0:68,:]))
        self.nonAbusedNonPTSDCortisol_rearr = np.vstack((self.nonAbusedNonPTSDCortisol[68:,:], self.nonAbusedNonPTSDCortisol[0:68,:]))

        # create the smoothed arrays
        self.abusedPTSDCortisol_smooth = self.abusedPTSDCortisol
        self.abusedPTSDCortisol_rearr_smooth = self.abusedPTSDCortisol_rearr
        self.nonAbusedPTSDCortisol_smooth = self.nonAbusedPTSDCortisol
        self.nonAbusedPTSDCortisol_rearr_smooth = self.nonAbusedPTSDCortisol_rearr
        self.nonAbusedNonPTSDCortisol_smooth = self.nonAbusedNonPTSDCortisol
        self.nonAbusedNonPTSDCortisol_rearr_smooth = self.nonAbusedNonPTSDCortisol_rearr

        self.abusedPTSDCortisol_smooth[2:-2,1] = self.__smoothing(self.abusedPTSDCortisol[:,1])
        self.abusedPTSDCortisol_rearr_smooth[2:-2,1] = self.__smoothing(self.abusedPTSDCortisol_rearr[:,1])
        self.nonAbusedPTSDCortisol_smooth[2:-2,1] = self.__smoothing(self.nonAbusedPTSDCortisol[:,1])
        self.nonAbusedPTSDCortisol_rearr_smooth[2:-2,1] = self.__smoothing(self.nonAbusedPTSDCortisol_rearr[:,1])
        self.nonAbusedNonPTSDCortisol_smooth[2:-2,1] = self.__smoothing(self.nonAbusedNonPTSDCortisol[:,1])
        self.nonAbusedNonPTSDCortisol_rearr_smooth[2:-2,1] = self.__smoothing(self.nonAbusedNonPTSDCortisol_rearr[:,1])

        # re-run the genfromtxt() commands because the smoothing overwrites the non-smoothed
        #  arrays for some reason
        self.abusedPTSDCortisol = np.genfromtxt("VeVaPy/data_files/Bremner-2007-abused-PTSD-cortisol.txt", dtype = float)
        self.nonAbusedPTSDCortisol = np.genfromtxt("VeVaPy/data_files/Bremner-2007-non-abused-PTSD-cortisol.txt", dtype = float)
        self.nonAbusedNonPTSDCortisol = np.genfromtxt("VeVaPy/data_files/Bremner-2007-non-abused-non-PTSD-cortisol.txt", dtype = float)

        # rearrange the data so that we start at 10AM like the Yehuda data
        self.abusedPTSDCortisol_rearr = np.vstack((self.abusedPTSDCortisol[68:,:], self.abusedPTSDCortisol[0:68,:]))
        self.nonAbusedPTSDCortisol_rearr = np.vstack((self.nonAbusedPTSDCortisol[68:,:], self.nonAbusedPTSDCortisol[0:68,:]))
        self.nonAbusedNonPTSDCortisol_rearr = np.vstack((self.nonAbusedNonPTSDCortisol[68:,:], self.nonAbusedNonPTSDCortisol[0:68,:]))

        # change the time steps of the rearranged arrays so that we start at 0
        for i in range(len(self.abusedPTSDCortisol[:,0])):
            self.abusedPTSDCortisol_rearr[i,0] = self.abusedPTSDCortisol[i,0]
            self.abusedPTSDCortisol_rearr_smooth[i,0] = self.abusedPTSDCortisol[i,0]
            self.nonAbusedPTSDCortisol_rearr[i,0] = self.nonAbusedPTSDCortisol[i,0]
            self.nonAbusedPTSDCortisol_rearr_smooth[i,0] = self.nonAbusedPTSDCortisol[i,0]
            self.nonAbusedNonPTSDCortisol_rearr[i,0] = self.nonAbusedNonPTSDCortisol[i,0]
            self.nonAbusedNonPTSDCortisol_rearr_smooth[i,0] = self.nonAbusedNonPTSDCortisol[i,0]

        if scale.lower() == "minutes":
            self.abusedPTSDCortisol[:,0] = np.multiply(self.abusedPTSDCortisol[:,0], 60)
            self.nonAbusedPTSDCortisol[:,0] = np.multiply(self.nonAbusedPTSDCortisol[:,0], 60)
            self.nonAbusedNonPTSDCortisol[:,0] = np.multiply(self.nonAbusedNonPTSDCortisol[:,0], 60)

            self.abusedPTSDCortisol_smooth[:,0] = np.multiply(self.abusedPTSDCortisol_smooth[:,0], 60)
            self.nonAbusedPTSDCortisol_smooth[:,0] = np.multiply(self.nonAbusedPTSDCortisol_smooth[:,0], 60)
            self.nonAbusedNonPTSDCortisol_smooth[:,0] = np.multiply(self.nonAbusedNonPTSDCortisol_smooth[:,0], 60)

            self.abusedPTSDCortisol_rearr[:,0] = np.multiply(self.abusedPTSDCortisol_rearr[:,0], 60)
            self.nonAbusedPTSDCortisol_rearr[:,0] = np.multiply(self.nonAbusedPTSDCortisol_rearr[:,0], 60)
            self.nonAbusedNonPTSDCortisol_rearr[:,0] = np.multiply(self.nonAbusedNonPTSDCortisol_rearr[:,0], 60)

            self.abusedPTSDCortisol_rearr_smooth[:,0] = np.multiply(self.abusedPTSDCortisol_rearr_smooth[:,0], 60)
            self.nonAbusedPTSDCortisol_rearr_smooth[:,0] = np.multiply(self.nonAbusedPTSDCortisol_rearr_smooth[:,0], 60)
            self.nonAbusedNonPTSDCortisol_rearr_smooth[:,0] = np.multiply(self.nonAbusedNonPTSDCortisol_rearr_smooth[:,0], 60)

    def nelson(self, scale):
        """
        The data in these sets is over a 2 hour and 20 minute period, and the
        patients underwent Trier Social Stress Tests at the 30 minute mark. As
        such, the data for most patients contains a pronounced spike in ACTH &
        CORT at that point and then returns to baseline afterwards.

        Contains mean ACTH & CORT data for melancholic MDD, atypical MDD,
        uncategorized MDD and healthy control subjects in the following arrays:
            - melancholicCortisol, melancholicACTH
            - atypicalCortisol, atypicalACTH
            - uncategorizedCortisol, uncategorizedACTH
            - healthyCortisol, healthyACTH

        Also contains data for each individual patient in the following arrays:
            - cortisol
            - ACTH

        The first column of Cortisol and ACTH is the time values, the second
        column is the mean values of all patients, and each subsequent column
        is an individual patient. There are 58 patients in the data set.

        For example:
            cortisol[11] is Patient ID 10's cortisol concentrations
            ACTH[59] is Patient ID 58's ACTH concentrations.
        """

        __ACTH_data = np.genfromtxt("VeVaPy/data_files/tsst_acth_nelson.txt")
        __cortisol_data = np.genfromtxt("VeVaPy/data_files/tsst_cort_nelson.txt")
        __subtypes = np.genfromtxt("VeVaPy/data_files/nelson-MDD-subtypes.txt")

        __ACTH_mean = np.zeros(11)
        __cortisol_mean = np.zeros(11)
        self.ACTH = np.zeros((11,60))
        self.cortisol = np.zeros((11,60))

        # compute the mean of all patients' ACTH and cortisol concentrations at each
        #  data point
        for i in range(len(__ACTH_data[1,:])-1):
            __ACTH_mean[i] = np.mean(__ACTH_data[:,i+1])
            __cortisol_mean[i] = np.mean(__cortisol_data[:,i+1])

        # create an array of the time points we have concentrations for in minutes
        __t_nelson = np.array([0, 15, 30, 40, 50, 65, 80, 95, 110, 125, 140])

        # put the time points in the first column of our ACTH and cortisol arrays
        #  and the mean of all patients' concentrations in the second column
        for i in range(len(__t_nelson)):
            self.ACTH[i,0] = __t_nelson[i]
            self.ACTH[i,1] = __ACTH_mean[i]

            self.cortisol[i,0] = __t_nelson[i]
            self.cortisol[i,1] = __cortisol_mean[i]

        # in the remaining columns, we put the concentrations at each data point for
        #  all remaining patients (one patient per column)
        for i in range(len(__ACTH_data)):
            for j in range(len(__t_nelson)):
                self.ACTH[j,i+2] = __ACTH_data[i,j+1]
                self.cortisol[j,i+2] = __cortisol_data[i,j+1]

        # Make lists of the indices in the ACTH and CORT arrays at which each
        #  subtype of patients are found
        __atypical_indices = []
        __melancholic_indices = []
        __neither_indices = []
        __healthy_indices = []

        for index, item in enumerate(__subtypes[:,1]):
            if item == 1:
                __atypical_indices.append(index)
            elif item == 2:
                __melancholic_indices.append(index)
            elif item == 3:
                __neither_indices.append(index)
            elif item == 4:
                __healthy_indices.append(index)

        # create lists of the indices in the ACTH and CORT arrays for each subtype
        #  (we need to shift by 2 columns, because of the time column and mean
        #  concentrations columns), and the patient IDs of the patients in each
        #  subtype
        __atypical_ids = []
        __melancholic_ids = []
        __neither_ids = []
        __healthy_ids = []
        for idx, item in enumerate(__atypical_indices):
            __atypical_indices[idx] += 2
            __atypical_ids.append(__subtypes[item,0])
        for idx, item in enumerate(__melancholic_indices):
            __melancholic_indices[idx] += 2
            __melancholic_ids.append(__subtypes[item,0])
        for idx, item in enumerate(__neither_indices):
            __neither_indices[idx] += 2
            __neither_ids.append(__subtypes[item,0])
        for idx, item in enumerate(__healthy_indices):
            __healthy_indices[idx] += 2
            __healthy_ids.append(__subtypes[item,0])

        # create lists of the patients' data arrays for ACTH & CORT for each subtype

        # create lists of arrays of atypical patient CORT and ACTH data
        __atypicalCortisolList = []
        __atypicalACTHList = []
        for idx in __atypical_indices:
            __atypicalCortisolList.append(self.cortisol[:,idx])
            __atypicalACTHList.append(self.ACTH[:,idx])

        # combine the lists of arrays into 2d arrays and transpose them so that
        #  each patients' data is in a column
        self.atypicalCortisol = np.vstack(__atypicalCortisolList)
        self.atypicalCortisol = np.transpose(self.atypicalCortisol)
        self.atypicalCortisol = np.insert(self.atypicalCortisol, 0, self.cortisol[:,0], axis = 1) # insert time in first column

        self.atypicalACTH = np.vstack(__atypicalACTHList)
        self.atypicalACTH = np.transpose(self.atypicalACTH)
        self.atypicalACTH = np.insert(self.atypicalACTH, 0, self.ACTH[:,0], axis = 1) # insert time in first column

        # create lists of arrays of melancholic patient CORT and ACTH data
        __melancholicCortisolList = []
        __melancholicACTHList = []
        for idx in __melancholic_indices:
            __melancholicCortisolList.append(self.cortisol[:,idx])
            __melancholicACTHList.append(self.ACTH[:,idx])

        # combine the lists of arrays into 2d arrays and transpose them so that
        #  each patients' data is in a column
        self.melancholicCortisol = np.vstack(__melancholicCortisolList)
        self.melancholicCortisol = np.transpose(self.melancholicCortisol)
        self.melancholicCortisol = np.insert(self.melancholicCortisol, 0, self.cortisol[:,0], axis = 1) # insert time in first column

        self.melancholicACTH = np.vstack(__melancholicACTHList)
        self.melancholicACTH = np.transpose(self.melancholicACTH)
        self.melancholicACTH = np.insert(self.melancholicACTH, 0, self.ACTH[:,0], axis = 1) # insert time in first column

        # create lists of arrays of neither atypical nor melancholic MDD patient CORT and ACTH data
        __neitherCortisolList = []
        __neitherACTHList = []
        for idx in __neither_indices:
            __neitherCortisolList.append(self.cortisol[:,idx])
            __neitherACTHList.append(self.ACTH[:,idx])

        # combine the lists of arrays into 2d arrays and transpose them so that
        #  each patients' data is in a column
        self.neitherCortisol = np.vstack(__neitherCortisolList)
        self.neitherCortisol = np.transpose(self.neitherCortisol)
        self.neitherCortisol = np.insert(self.neitherCortisol, 0, self.cortisol[:,0], axis = 1) # insert time in first column

        self.neitherACTH = np.vstack(__neitherACTHList)
        self.neitherACTH = np.transpose(self.neitherACTH)
        self.neitherACTH = np.insert(self.neitherACTH, 0, self.ACTH[:,0], axis = 1) # insert time in first column

        # create lists of arrays of healthy patient CORT and ACTH data
        __healthyCortisolList = []
        __healthyACTHList = []
        for idx in __healthy_indices:
            __healthyCortisolList.append(self.cortisol[:,idx])
            __healthyACTHList.append(self.ACTH[:,idx])

        # combine the lists of arrays into 2d arrays and transpose them so that
        #  each patients' data is in a column
        self.healthyCortisol = np.vstack(__healthyCortisolList)
        self.healthyCortisol = np.transpose(self.healthyCortisol)
        self.healthyCortisol = np.insert(self.healthyCortisol, 0, self.cortisol[:,0], axis = 1) # insert time in first column

        self.healthyACTH = np.vstack(__healthyACTHList)
        self.healthyACTH = np.transpose(self.healthyACTH)
        self.healthyACTH = np.insert(self.healthyACTH, 0, self.ACTH[:,0], axis = 1) # insert time in first column

        # compute means of all patients in each subtype of depression, all control patients,
        # and all subtypes of depression combined

        # initialize empty arrays to fill with mean values
        self.healthyCortisol_mean = np.zeros((len(self.healthyCortisol[:,0]), 1))
        self.healthyACTH_mean = np.zeros((len(self.healthyACTH[:,0]), 1))
        self.depressedCortisol_mean = np.zeros((len(self.melancholicCortisol[:,0]), 1))
        self.depressedACTH_mean = np.zeros((len(self.melancholicACTH[:,0]), 1))
        self.melancholicCortisol_mean = np.zeros((len(self.melancholicCortisol[:,0]), 1))
        self.melancholicACTH_mean = np.zeros((len(self.melancholicACTH[:,0]), 1))
        self.atypicalCortisol_mean = np.zeros((len(self.atypicalCortisol[:,0]), 1))
        self.atypicalACTH_mean = np.zeros((len(self.atypicalACTH[:,0]), 1))
        self.neitherCortisol_mean = np.zeros((len(self.neitherCortisol[:,0]), 1))
        self.neitherACTH_mean = np.zeros((len(self.neitherACTH[:,0]), 1))

        # loop through each time step and add the sum of all patients' values at that time step
        for i in range(len(self.healthyCortisol[:,0])):
            self.healthyCortisol_mean[i] += np.sum(self.healthyCortisol[i, 1:])
            self.depressedCortisol_mean[i] += np.sum(self.melancholicCortisol[i, 1:])
            self.melancholicCortisol_mean[i] += np.sum(self.melancholicCortisol[i, 1:])
            self.depressedCortisol_mean[i] += np.sum(self.atypicalCortisol[i, 1:])
            self.atypicalCortisol_mean[i] += np.sum(self.atypicalCortisol[i, 1:])
            self.depressedCortisol_mean[i] += np.sum(self.neitherCortisol[i, 1:])
            self.neitherCortisol_mean[i] += np.sum(self.neitherCortisol[i, 1:])

            self.healthyACTH_mean[i] += np.sum(self.healthyACTH[i, 1:])
            self.depressedACTH_mean[i] += np.sum(self.melancholicACTH[i, 1:])
            self.melancholicACTH_mean[i] += np.sum(self.melancholicACTH[i, 1:])
            self.depressedACTH_mean[i] += np.sum(self.atypicalACTH[i, 1:])
            self.atypicalACTH_mean[i] += np.sum(self.atypicalACTH[i, 1:])
            self.depressedACTH_mean[i] += np.sum(self.neitherACTH[i, 1:])
            self.neitherACTH_mean[i] += np.sum(self.neitherACTH[i, 1:])

        # divide by the number of patients for each set to get the mean
        self.healthyCortisol_mean = self.healthyCortisol_mean/15
        self.healthyCortisol_mean = np.insert(self.healthyCortisol_mean, 0, self.cortisol[:,0], axis = 1) # insert time in first column
        self.healthyACTH_mean = self.healthyACTH_mean/15
        self.healthyACTH_mean = np.insert(self.healthyACTH_mean, 0, self.ACTH[:,0], axis = 1) # insert time in first column

        self.depressedCortisol_mean = self.depressedCortisol_mean/43
        self.depressedCortisol_mean = np.insert(self.depressedCortisol_mean, 0, self.cortisol[:,0], axis = 1) # insert time in first column
        self.depressedACTH_mean = self.depressedACTH_mean/43
        self.depressedACTH_mean = np.insert(self.depressedACTH_mean, 0, self.ACTH[:,0], axis = 1) # insert time in first column

        self.melancholicCortisol_mean = self.melancholicCortisol_mean/15
        self.melancholicCortisol_mean = np.insert(self.melancholicCortisol_mean, 0, self.cortisol[:,0], axis = 1) # insert time in first column
        self.melancholicACTH_mean = self.melancholicACTH_mean/15
        self.melancholicACTH_mean = np.insert(self.melancholicACTH_mean, 0, self.ACTH[:,0], axis = 1) # insert time in first column

        self.atypicalCortisol_mean = self.atypicalCortisol_mean/14
        self.atypicalCortisol_mean = np.insert(self.atypicalCortisol_mean, 0, self.cortisol[:,0], axis = 1) # insert time in first column
        self.atypicalACTH_mean = self.atypicalACTH_mean/14
        self.atypicalACTH_mean = np.insert(self.atypicalACTH_mean, 0, self.ACTH[:,0], axis = 1) # insert time in first column

        self.neitherCortisol_mean = self.neitherCortisol_mean/14
        self.neitherCortisol_mean = np.insert(self.neitherCortisol_mean, 0, self.cortisol[:,0], axis = 1) # insert time in first column
        self.neitherACTH_mean = self.neitherACTH_mean/14
        self.neitherACTH_mean = np.insert(self.neitherACTH_mean, 0, self.ACTH[:,0], axis = 1) # insert time in first column

        if scale.lower() == "hours":
            self.cortisol[:,0] = np.divide(self.cortisol[:,0], 60)
            self.ACTH[:,0] = np.divide(self.ACTH[:,0], 60)
            self.depressedCortisol_mean[:,0] = np.divide(self.depressedCortisol_mean[:,0], 60)
            self.depressedACTH_mean[:,0] = np.divide(self.depressedACTH_mean[:,0], 60)
            self.atypicalCortisol[:,0] = np.divide(self.atypicalCortisol[:,0], 60)
            self.atypicalCortisol_mean[:,0] = np.divide(self.atypicalCortisol_mean[:,0], 60)
            self.atypicalACTH[:,0] = np.divide(self.atypicalACTH[:,0], 60)
            self.atypicalACTH_mean[:,0] = np.divide(self.atypicalACTH_mean[:,0], 60)
            self.melancholicCortisol[:,0] = np.divide(self.melancholicCortisol[:,0], 60)
            self.melancholicCortisol_mean[:,0] = np.divide(self.melancholicCortisol_mean[:,0], 60)
            self.melancholicACTH[:,0] = np.divide(self.melancholicACTH[:,0], 60)
            self.melancholicACTH_mean[:,0] = np.divide(self.melancholicACTH_mean[:,0], 60)
            self.neitherCortisol[:,0] = np.divide(self.neitherCortisol[:,0], 60)
            self.neitherCortisol_mean[:,0] = np.divide(self.neitherCortisol_mean[:,0], 60)
            self.neitherACTH[:,0] = np.divide(self.neitherACTH[:,0], 60)
            self.neitherACTH_mean[:,0] = np.divide(self.neitherACTH_mean[:,0], 60)
            self.healthyCortisol[:,0] = np.divide(self.healthyCortisol[:,0], 60)
            self.healthyCortisol_mean[:,0] = np.divide(self.healthyCortisol_mean[:,0], 60)
            self.healthyACTH[:,0] = np.divide(self.healthyACTH[:,0], 60)
            self.healthyACTH_mean[:,0] = np.divide(self.healthyACTH_mean[:,0], 60)

    def patientF(self, scale):
        """
        Import the data for "Patient F" from the paper by Bangsgaard & Ottesen
        (2017) and save the data to separate arrays for cortisol and ACTH
        concentration data.

        Contains cortisol and ACTH data in the following arrays:
            - cortisol
            - ACTH

        Also create smoothed versions of these arrays with _smooth at the end
        of the variable name.

        For example:
            cortisol_smooth
        """
        self.cortisol = np.genfromtxt("VeVaPy/data_files/Bangsgaard-Ottesen-2017-patient-f-cortisol-data.txt", dtype = float)
        self.ACTH = np.genfromtxt("VeVaPy/data_files/Bangsgaard-Ottesen-2017-patient-f-ACTH-data.txt", dtype = float)

        self.cortisol_smooth = self.cortisol
        self.ACTH_smooth = self.ACTH

        self.cortisol_smooth[2:-2,1] = self.__smoothing(self.cortisol[:,1])
        self.ACTH_smooth[2:-2,1] = self.__smoothing(self.ACTH[:,1])

        self.cortisol = np.genfromtxt("VeVaPy/data_files/Bangsgaard-Ottesen-2017-patient-f-cortisol-data.txt", dtype = float)
        self.ACTH = np.genfromtxt("VeVaPy/data_files/Bangsgaard-Ottesen-2017-patient-f-ACTH-data.txt", dtype = float)

        if scale.lower() == "minutes":
            self.cortisol[:,0] = np.multiply(self.cortisol[:,0], 60)
            self.ACTH[:,0] = np.multiply(self.ACTH[:,0], 60)

            self.cortisol_smooth[:,0] = np.multiply(self.cortisol_smooth[:,0], 60)
            self.ACTH_smooth[:,0] = np.multiply(self.ACTH_smooth[:,0], 60)



# File Name: ODEsolver.py
# Description: this file will contain a function that performs the ODE solver
#   steps to improve readability in the model notebooks
# Author: Christopher Parker
# Created: Mon Jan 24, 2022 | 03:39P EST
# Last Modified: Mon Sep 12, 2022 | 02:08P EDT

# Importing through __init__.py now
#import scipy.integrate as sci
#import numpy as np

# This function takes an ODE system (defined as a function), initial conditions,
#  the time interval for integration. It can also (optionally) accept values for
#  all of the ODE solver options and delays on CRH, ACTH and/or cortisol
def solve(ode_system, ics, t_start, t_step, t_end, ode_steps = 100000, ode_atol = 3e-12, ode_rtol = 1e-12, y_index0 = 0, y_index1 = 1, y_index2 = 2, tau0 = 0, tau1 = 0, tau2 = 0, delay = [False, False, False], delay_rough = False):

    # If any of the delay flags has been set to True, define global variables 
    #  for delayed CRH, ACTH and CORT values.
    if delay[0] or delay[1] or delay[2]:
        global delayedCRH, delayedACTH, delayedCORT
        delayedCRH = ics[y_index0]
        delayedACTH = ics[y_index1]
        delayedCORT = ics[y_index2]

        # These variables will be used to hold the previous index in the time
        #  array at which we found the delayed values of ACTH and CORT, resp.
        t_index0 = 0
        t_index1 = 0
        t_index2 = 0

    # Initialize empty arrays for storing values of ts and ys (the time steps
    #  and y-values of the solution to the ODE system)
    ts = []
    ys = []

    # These lines define the ODE solver object using the arguments passed to 
    #  solve(), or the defaults if options haven't been passed
    solver = sci.ode(ode_system)
    solver.set_integrator('lsoda', method='bdf', atol = ode_atol, rtol = ode_rtol, nsteps=ode_steps)
    solver.set_initial_value(ics, t_start)

    # Here is where we loop through the time interval of integration, but only
    #  as long as solver.successful() returns True--which means that we haven't
    #  had the solver fail to solve any steps
    while solver.successful() and solver.t < t_end:
        # If the user sets the index 0 delay flag to True, we compute the delayed 
        #  value of CRH
        # We also check whether the delay_rough flag is set, and depending on this
        #  we either run the normal version of the delay functions or the _rough
        #  version
        if delay[0] and delay_rough:
            delayedCRH, t_index0 = delay_CRH_rough(solver.t, tau0, ts, ys, ics, t_index0, y_index0)
        if delay[0] and (not delay_rough):
            delayedCRH, t_index0 = delay_CRH(solver.t, tau0, ts, ys, ics, t_index0, y_index0)

        # If the user sets the index 1 delay flag to True, we compute
        #  the delayed value of ACTH
        # We also check the delay_rough flag again
        if delay[1] and delay_rough:
            delayedACTH, t_index1 = delay_ACTH_rough(solver.t, tau1, ts, ys, ics, t_index1, y_index1)
        if delay[1] and (not delay_rough):
            delayedACTH, t_index1 = delay_ACTH(solver.t, tau1, ts, ys, ics, t_index1, y_index1)

        # If the user sets the index 2 delay flag to True, we compute the delayed
        #  value of CORT
        # We also check the delay_rough flag again
        if delay[2] and delay_rough:
            delayedCORT, t_index2 = delay_CORT_rough(solver.t, tau2, ts, ys, ics, t_index2, y_index2)
        if delay[2] and (not delay_rough):
            delayedCORT, t_index2 = delay_CORT(solver.t, tau2, ts, ys, ics, t_index2, y_index2)

        # This is where we actually run the solver (telling it to use the last
        #  time it solved plus the step size we set) and put the time step and
        #  y-value it returns into the arrays ts and ys, respectively
        solver.integrate(solver.t + t_step)
        ts.append(solver.t)
        ys.append(solver.y)

    # Reshape the output from sci.ode to an array with the times in the first
    #  column
    ts = np.reshape(ts, (len(ts), 1))
    ys = np.vstack(ys)
    timeSeries = np.hstack((ts, ys))

    return timeSeries

# This function takes the solutions to all of the previously solved time steps
#  and the time of the current step, and determines what the solution for ACTH
#  concentration was tau1 time units ago (depends on time scale, usually
#  minutes or hours)
def delay_ACTH(t, tau1, ts, ys, ics, t_index1, y_index1):
    # t_round is equal to t rounded down to the nearest multiple of 0.01.
    # This is necessary because as the ODE solver is running, it uses much 
    #  smaller steps than 0.01 at times (especially when the system is 
    #  very stiff).
    # The uneven size of time steps used by the ODE solver makes this entire
    #  process much more complex than it would otherwise be, as we need to add
    #  in some rounding of time steps and allow for a slight error tolerance
    #  if we cannot find the exact time step we are searching for.
    t_round = t - t%0.01

    # Generate an array of values near to t_round - tau1, in case we skipped 
    #  the exact time step.
    t_tau1 = np.arange(t_round - tau1, t_round - tau1 + 0.01, 0.01)

    # First, we check if t_round is greater than tau1 (in other words, has the
    #  system been solved for longer than tau1 time units?)
    # If it isn't, we will just use the initial time point of y[1] until it is.
    if t_round > tau1:
        # This for statement iterates through the array of previously solved 
        #  time steps (saved in the array ts).
        for ts_index, ts_item in enumerate(ts[t_index1:]):
            # First, try to match the exact time step t - tau1.
            if (ts_item - ts_item%0.01) == (t_round - tau1 - (t_round - tau1)%0.01):
                # Now that we have found the correct time step for the delay, 
                #  we set a variable delayedACTH to the value of y[1] from the delayed 
                #  time step, and set t_index1 to the index in ts at which we 
                #  found it.
                delayedACTH = ys[ts_index + t_index1][y_index1]
                t_index1 = ts_index + t_index1

                # end the for loop now that we have found a match
                break
        # If we didn't find an exact match, try finding a match within 
        #  0.01 minutes of the exact time.
        else:
            for ts_index, ts_item in enumerate(ts[t_index1:]):
                # Let's try using a double for loop so that we can round both 
                #  the value from ts and the value of t - tau1 because we seem 
                #  to be suffering from lots of rounding error because 
                #  computers can't handle floats exactly.
                for ttau1_item in t_tau1:
                    # Check if we have a solved time step within 0.01 minutes 
                    #  of t - tau1 (we may not have an exact match, it seems to 
                    #  iterate more than 0.01 in a step at times).
                    # Note that we round ts_item and ttau1_item down to the 
                    #  nearest multiple of 0.01 because of rounding error.
                    if (ts_item - ts_item%0.01) == (ttau1_item - ttau1_item%0.01):
                        # Now that we have found the correct time step for the 
                        #  delay, we set a variable delayedACTH to the value of y[1] 
                        #  from the delayed time step, and set t_index1 to the 
                        #  index in ts at which we found it.

                        delayedACTH = ys[ts_index + t_index1][y_index1]
                        t_index1 = ts_index + t_index1

                        # End the for loop early now that we have found a match.
                        break
                # This is a useful bit of code to break out of a double for 
                #  loop if we succeeded in doing what we wanted. If the code 
                #  makes it to the break inside the if statement, we do not run 
                #  this else statement and instead break out of the outer for 
                #  loop. If we do not make it to the break inside the if 
                #  statement, we run the else statement and continue to the 
                #  next step of the outer for loop, skipping the break below.
                else:
                    continue

                break

            # Same idea as explained above, if we don't break out of the for 
            #  loop, run this code.
            # If we are running this, it means we didn't find a match 
            #  within 0.01 of t - tau1.
            else:
                # Expand the range of values around the delayed time step we 
                #  are willing to accept.
                #t_tau1 = np.arange(t_round - tau1 - 0.1, t_round - tau1 + 0.1, 0.01)
                t_tau1 = np.arange(t_round - tau1, t_round - tau1 + 0.1, 0.01)

                # Now we simply run the same for loop as above.
                for ts_index, ts_item in enumerate(ts[t_index1:]):
                    for ttau1_item in t_tau1:
                        if (ts_item - ts_item%0.01) == (ttau1_item - ttau1_item%0.01):
                            delayedACTH = ys[ts_index + t_index1][y_index1]
                            t_index1 = ts_index + t_index1

                            break
                    else:
                        continue
                    break

                # If we don't have a match within 0.1 minutes, set delayedACTH 
                #  to the initial value for ACTH and tell the user that there's
                #  a problem.
                #
                # Comment this code out if expanding the search range further 
                #  by uncommenting the code below.
                else:
                    delayedACTH = ics[y_index1]
                    print("No match for y[{}] delayed by tau1 within 0.1 minutes".format(y_index1))

    # If t_round <= tau1, just use the initial value for y[1]
    else:
        delayedACTH = ics[y_index1]

    return delayedACTH, t_index1

# In case any model needs to use delayed values for CRH, we will include a 
#  function here to find those--it will be just like the ACTH delay function
def delay_CRH(t, tau0, ts, ys, ics, t_index0, y_index0):
    t_round = t - t%0.01
    t_tau0 = np.arange(t_round - tau0, t_round - tau0 + 0.01, 0.01)

    if t_round > tau0:
        for ts_index, ts_item in enumerate(ts[t_index0:]):
            if (ts_item - ts_item%0.01) == (t_round - tau0 - (t_round - tau0)%0.01):
                delayedCRH = ys[ts_index + t_index0][y_index0]
                t_index0 = ts_index + t_index0
                break
        else:
            for ts_index, ts_item in enumerate(ts[t_index0:]):
                for ttau0_item in t_tau0:
                    if (ts_item - ts_item%0.01) == (ttau0_item - ttau0_item%0.01):
                        delayedCRH = ys[ts_index + t_index0][y_index0]
                        t_index0 = ts_index + t_index0
                        break
                else:
                    continue
                break
            else:
                t_tau0 = np.arange(t_round - tau0, t_round - tau0 + .1, 0.01)
                for ts_index, ts_item in enumerate(ts[t_index0:]):
                    for ttau0_item in t_tau0:
                        if (ts_item - ts_item%0.01) == (ttau0_item - ttau0_item%0.01):
                            delayedCRH = ys[ts_index + t_index0][y_index0]
                            t_index0 = ts_index + t_index0
                            break
                    else:
                        continue
                    break
                else:
                    delayedCRH = ics[y_index0]
                    print("No match for y[{}] delayed by tau0 within .1 minute.".format(y_index0))
    else:
        delayedCRH = ics[y_index0]

    return delayedCRH

# Same as delay_ACTH, but for tau2 now (which is a longer delay) and now we are 
#  concerned with setting y[2], since the tau2 delay is the delay in 
#  cortisol action.
def delay_CORT(t, tau2, ts, ys, ics, t_index2, y_index2):
    t_round = t - t%0.01
    #t_tau2 = np.arange(t_round - tau2 - 0.01, t_round - tau2 + 0.01, 0.01)
    t_tau2 = np.arange(t_round - tau2, t_round - tau2 + 0.01, 0.01)

    if t_round > tau2:
        for ts_index, ts_item in enumerate(ts[t_index2:]):
            if (ts_item - ts_item%0.01) == (t_round - tau2 - (t_round - tau2)%0.01):
                delayedCORT = ys[ts_index + t_index2][y_index2]
                t_index2 = ts_index + t_index2
                break
        else:
            for ts_index, ts_item in enumerate(ts[t_index2:]):
                for ttau2_item in t_tau2:
                    if (ts_item - ts_item%0.01) == (ttau2_item - ttau2_item%0.01):
                        delayedCORT = ys[ts_index + t_index2][y_index2]
                        t_index2 = ts_index + t_index2
                        break
                else:
                    continue
                break
            else:
                #t_tau2 = np.arange(t_round - tau2 - 0.1, t_round - tau2 + 0.1, 0.01)
                t_tau2 = np.arange(t_round - tau2, t_round - tau2 + 0.1, 0.01)
                for ts_index, ts_item in enumerate(ts[t_index2:]):
                    for ttau2_item in t_tau2:
                        if (ts_item - ts_item%0.01) == (ttau2_item - ttau2_item%0.01):
                            delayedCORT = ys[ts_index + t_index2][y_index2]
                            t_index2 = ts_index + t_index2
                            break
                    else:
                        continue
                    break
                else:
                    delayedCORT = ics[y_index2]
                    print("No match for y[{}] delayed by tau2 within 0.1 minutes".format(y_index2))

    # if t_round <= tau2, just use the initial value for y[2]
    else:
        delayedCORT = ics[y_index2]

    return delayedCORT, t_index2

# These functions check fewer time steps around t_round - tau1 in order to improve
#  runtime. Other than using %0.5 instead of %0.01, the functions are identical.
def delay_ACTH_rough(t, tau1, ts, ys, ics, t_index1, y_index1):
    t_round = t - t%0.5
    t_tau1 = np.arange(t_round - tau1, t_round - tau1 + 0.5, 0.5)

    if t_round > tau1:
        for ts_index, ts_item in enumerate(ts[t_index1:]):
            if (ts_item - ts_item%0.5) == (t_round - tau1 - (t_round - tau1)%0.5):
                delayedACTH = ys[ts_index + t_index1][y_index1]
                t_index1 = ts_index + t_index1
                break
        else:
            for ts_index, ts_item in enumerate(ts[t_index1:]):
                for ttau1_item in t_tau1:
                    if (ts_item - ts_item%0.5) == (ttau1_item - ttau1_item%0.5):
                        delayedACTH = ys[ts_index + t_index1][y_index1]
                        t_index1 = ts_index + t_index1
                        break
                else:
                    continue
                break
            else:
                t_tau1 = np.arange(t_round - tau1, t_round - tau1 + 1, 0.5)

                for ts_index, ts_item in enumerate(ts[t_index1:]):
                    for ttau1_item in t_tau1:
                        if (ts_item - ts_item%0.5) == (ttau1_item - ttau1_item%0.5):
                            delayedACTH = ys[ts_index + t_index1][y_index1]
                            t_index1 = ts_index + t_index1
                            break
                    else:
                        continue
                    break
                else:
                    delayedACTH = ics[y_index1]
                    print("No match for y[{}] delayed by tau1 within 1 minute.".format(y_index1))
    else:
        delayedACTH = ics[y_index1]

    return delayedACTH, t_index1

def delay_CRH_rough(t, tau0, ts, ys, ics, t_index0, y_index0):
    t_round = t - t%0.5
    t_tau0 = np.arange(t_round - tau0, t_round - tau0 + 0.5, 0.5)

    if t_round > tau0:
        for ts_index, ts_item in enumerate(ts[t_index0:]):
            if (ts_item - ts_item%0.5) == (t_round - tau0 - (t_round - tau0)%0.5):
                delayedCRH = ys[ts_index + t_index0][y_index0]
                t_index0 = ts_index + t_index0
                break
        else:
            for ts_index, ts_item in enumerate(ts[t_index0:]):
                for ttau0_item in t_tau0:
                    if (ts_item - ts_item%0.5) == (ttau0_item - ttau0_item%0.5):
                        delayedCRH = ys[ts_index + t_index0][y_index0]
                        t_index0 = ts_index + t_index0
                        break
                else:
                    continue
                break
            else:
                t_tau0 = np.arange(t_round - tau0, t_round - tau0 + 1, 0.5)
                for ts_index, ts_item in enumerate(ts[t_index0:]):
                    for ttau0_item in t_tau0:
                        if (ts_item - ts_item%0.5) == (ttau0_item - ttau0_item%0.5):
                            delayedCRH = ys[ts_index + t_index0][y_index0]
                            t_index0 = ts_index + t_index0
                            break
                    else:
                        continue
                    break
                else:
                    delayedCRH = ics[y_index0]
                    print("No match for y[{}] delayed by tau0 within 1 minute.".format(y_index0))
    else:
        delayedCRH = ics[y_index0]

    return delayedCRH, t_index0

def delay_CORT_rough(t, tau2, ts, ys, ics, t_index2, y_index2):
    t_round = t - t%0.5
    t_tau2 = np.arange(t_round - tau2, t_round - tau2 + 0.5, 0.5)

    if t_round > tau2:
        for ts_index, ts_item in enumerate(ts[t_index2:]):
            if (ts_item - ts_item%0.5) == (t_round - tau2 - (t_round - tau2)%0.5):
                delayedCORT = ys[ts_index + t_index2][y_index2]
                t_index2 = ts_index + t_index2
                break
        else:
            for ts_index, ts_item in enumerate(ts[t_index2:]):
                for ttau2_item in t_tau2:
                    if (ts_item - ts_item%0.5) == (ttau2_item - ttau2_item%0.5):
                        delayedCORT = ys[ts_index + t_index2][y_index2]
                        t_index2 = ts_index + t_index2
                        break
                else:
                    continue
                break
            else:
                t_tau2 = np.arange(t_round - tau2, t_round - tau2 + 1, 0.5)
                for ts_index, ts_item in enumerate(ts[t_index2:]):
                    for ttau2_item in t_tau2:
                        if (ts_item - ts_item%0.5) == (ttau2_item - ttau2_item%0.5):
                            delayedCORT = ys[ts_index + t_index2][y_index2]
                            t_index2 = ts_index + t_index2
                            break
                    else:
                        continue
                    break
                else:
                    delayedCORT = ics[y_index2]
                    print("No match for y[{}] delayed by tau2 within 1 minute.".format(y_index2))
    else:
        delayedCORT = ics[y_index2]

    return delayedCORT, t_index2



# File Name: visualize.py
# Description: Contains a class for creating graphs of model simulations,
#  with or without real-world data displayed, also
# Author: Christopher Parker
# Created: Sat Aug 27, 2022 | 04:25P EDT
# Last Modified: Mon Sep 12, 2022 | 02:13P EDT

# Importing via __init__.py now
#from IPython.display import clear_output
#from functools import wraps
#import numpy as np
#import matplotlib.pyplot as plt
#import re


class Visualizer:
    """
    This class will walk the user through the inputs necessary to plot simulation
    data against real-world data for the purposes of model verification and 
    validation. 

    Because the module exists in a separate namespace than the code using it,
    we unfortunately need to pass in the dictionary containing the variables
    defined in the global namespace for the main code. This will allow the
    user to type in strings to input boxes and have that retrieve the values
    saved in variables
    """
    def __init__(self, globals_input):
        # I'm going to have the class get a bunch of inputs from the user here, so they don't have to pass arguments
        # Basically, that's because it would take a lot of arguments to get all the information needed, and this will
        #  make things more legible, and easier to understand.

        # This loop structure will keep requesting that the user input the indices of the variables to graph until
        #  the input is correct (must be integers >= 0)
        self.globals_dict = globals_input
        while True:
            try:
                # Tell the user what input to enter and then save that input into the variable self.var_indices
                self.var_indices = input("Indices (in the array y0) of the variables you wish to plot (\033[1mNote that Python indexing starts at 0!\033[0m): ")

                # Convert the user input into a list of integers by separating at the commas
                self.var_indices = [int(i) for i in list(self.var_indices.split(','))]

                # Ensure that all list entries are zero or greater, now that we know they are integers
                for entry in self.var_indices:
                    if entry < 0:
                        raise ValueError

            # If a ValueError is raised by the input function receiving an invalid value, then we print an error message
            #  and start back at the beginning of the while True loop
            except ValueError:
                # We use this function to clear the output in the Jupyter notebook often, so that it doesn't get cluttered,
                #  and remains easy to read
                clear_output()
                # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                print("\033[1mInput must be an integer or a comma-separated list of integers, zero or greater and matching in number the input for number of graphs\033[0m")
                continue

            # Set a variable for the number of graphs requested, this will come in handy later
            self.num_graphs = len(self.var_indices)

            # We only break out of the while True loop if we don't go through except ValueError
            clear_output()
            break

        while True:
            try:
                # Ask the user for the name of the array containing data to plot
                self.array_name = input("Enter the name of the array to be graphed (simData_all for optimized simulations, data_no_opt for non-optimized)")
                # This line just checks if the name given by the user has been used somewhere in the code. If not,
                #  it will raise a KeyError, and we will try again.
                self.globals_dict[self.array_name]
            except KeyError:
                clear_output()
                # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                print("\033[1mMust enter the name of a globally defined array\033[0m")
                continue
            clear_output()
            break

        while True:
            try:
                # Ask the user how many iterations of parameter optimization are contained in the array
                #  (Prompt the user to answer 0 for no parameter optimization)
                # This will raise a ValueError for anything other than an integer
                self.num_iters = int(input(f"Number of iterations of parameter optimization contained in {self.array_name} (Enter 0 for no optimization): "))

                # If the number given is negative, raise a ValueError, also
                if self.num_iters < 0:
                    raise ValueError

                # Check that the array has enough columns for the number of variables to be graphed
                try:
                    num_cols = len(self.globals_dict[self.array_name][0,:])
                    if num_cols < (self.num_graphs+1)*self.num_iters:
                        raise TypeError
                except TypeError:
                    clear_output()
                    print("""The array given does not have the proper number of columns for the graphs requested.
            It must contain the time steps in the first column, and then a column for each variable being graphed
            (and this same structure must be repeated for each iteration of parameter optimization).\n""")
                    continue
            # If a ValueError is raised by the input function receiving an invalid value, then we print an error message
            #  and start back at the beginning of the while True loop
            except ValueError:
                clear_output()
                # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                print("\033[1mInput must be either a positive integer or zero.")
                continue
            # We only break out of the while True loop if we don't go through except ValueError
            clear_output()
            break

        while True:
            try:
                # Ask the user if they would like to include data from a real-world data set on the graphs, and the
                #  name of the data set if yes. Raises a ValueError if the input is not an integer
                self.real_data_flag = input("Graph real-world data along with simulations? (Y/N)")

                # If the input doesn't start with either y or n, we raise a ValueError. I'm only looking at the first
                #  letter to allow for typing "yes" or "no", also
                if self.real_data_flag[0].upper() != 'Y' and self.real_data_flag[0].upper() != 'N':
                    raise ValueError
            # If a ValueError is raised by the input function receiving an invalid value, then we print an error message
            #  and start back at the beginning of the while True loop
            except ValueError:
                clear_output()
                # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                print("\033[1mInput must be either Y or N.")
                continue

            # If the user answered in the affirmative, set the flag to True
            if self.real_data_flag[0].upper() == 'Y':
                self.real_data_flag = True

            # If the user answered in the negative, set the flag to False
            elif self.real_data_flag[0].upper() == 'N':
                self.real_data_flag = False

                # Since no real-world data has been requested, we will just set the array name for each graph to ''
                #  to ensure we don't get any errors
                self.real_data_arrays = ['']*self.num_graphs

            # We only break out of the while True loop if we don't go through except ValueError
            clear_output()
            break

        if self.real_data_flag:
            # Define an empty list to hold any real-world data array names we might need
            self.real_data_arrays = []
            self.real_data_objects = []
            self.real_data_objectArrays = []
            self.real_data_col_numbers = []
            # If the user has indicated that they want real-world data included on one or more graphs, we need to get the
            #  arrays in which that data is stored from them and which graphs it should go on
            for i in range(self.num_graphs):
                while True:
                    try:
                        # Ask the user for the name of the first array containing real-world data
                        temp_array_name = input(f"Enter the name of real-world data array to show on graph #{i+1} (leave blank for none) and the column number of the data \
                        (\033[1mseparated by a comma!\033[0m 1 for all sets other than Nelson data): ")

                        # If the user leaves the input box blank, we skip to the next for loop iteration
                        if temp_array_name == "":
                            # Set these to "" so that it's clear that both the global namespace and object namespace
                            #  arrays are undefined in this case
                            temp_data_obj = ""
                            temp_data_array = ""
                            temp_column_number = ""
                            clear_output()
                            break

                        try:
                            # Split the array name from the column number, and make sure the column number is an integer
                            #  greater than 0
                            temp_array_name, temp_column_number = temp_array_name.split(',')

                            # This is a simple regex substitution to remove any whitespace, just in case the user
                            #  put in any spaces. Luckily int() removes them by default, so we don't need to worry
                            #  for the column number
                            temp_array_name = re.sub("\s", "", temp_array_name)
                            temp_column_number = int(temp_column_number)

                            if temp_column_number <=0:
                                raise ValueError

                        except ValueError:
                            clear_output()
                            # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                            print("\033[1mRemember to enter the column number after the array name, separated by a comma\033[0m")
                            continue

                        # Check if the array is defined in the global namespace and make sure it has the proper number of
                        #  columns
                        self.globals_dict[temp_array_name][:,temp_column_number]

                        # If we make it to this statement, we know that the array given is in the global namespace,
                        #  not an object's namespace. So we will just define the object name variables as "", so that
                        #  while plotting we can check which type of array we are dealing with
                        temp_data_obj = ""
                        temp_data_array = ""

                    # If the array name is not defined in the global namespace, we need to check and see if it's an object
                    #  (such as one created from the dataImport module). These objects each have their own namespace, so
                    #  it takes a bit of extra manipulation to use them.
                    except KeyError:
                        # First, we try to split the name given by the user at a period (because an array contained
                        #  inside an instance of an object will have a name such as nelson.ACTH). The first string is the
                        #  name of an instance of an object, and the second is the array name that is hopefully contained in
                        #  the object.
                        try:
                            temp_data_obj, temp_data_array = temp_array_name.split('.')
                        # If there is no period, it will raise a ValueError, and we know that the user has input an invalid
                        #  array name, so we go back to the beginning of the while True loop
                        except ValueError:
                            clear_output()
                            # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                            print("\033[1mMust enter the name of a globally defined array\033[0m")
                            continue
                        # This try statement could give us one of three different errors. If the object name given
                        #  is invalid (it doesn't have its own namespace), then it will raise a TypeError. If the
                        #  object name is valid, but has no array with the given name, then it will raise a KeyError.
                        #  If both names are valid, but the array doesn't have enough columns,
                        #  it will raise an IndexError.
                        try:
                            # Parts of this statement:
                            #     - self.globals_dict[temp_data_obj] gets the value of the variable associated with the name
                            #        stored in the string temp_data_obj
                            #     - vars() is the function to access the dictionary containing variable names and values
                            #        for the namespace of the object passed as an argument
                            #     - So by passing self.globals_dict[temp_data_obj], we are accessing the dictionary of the object
                            #        namespace of the variable named temp_data_obj
                            #     - Then we want to access a variable from the dictionary, so we index it with [temp_data_array],
                            #        which returns the value of the variable of the same name as the string stored in temp_data_array
                            #     - Now that we have the variable stored in the object and array names passed by the user, we check
                            #        to see if the column defined by temp_column_number exists, by indexing with [:,temp_column_number]
                            #        to access all rows of data contained in that column
                            #
                            # This statement is unfortunately complex, but I believe it is the best way to handle accessing arrays
                            #  stored in objects at the request of the user. We use the same format in the make_graphs function to access
                            #  the user-requested arrays for plotting.
                            vars(self.globals_dict[temp_data_obj])[temp_data_array][:,temp_column_number]
                            # At this point, we know that the array is in an object's namespace, so we will define
                            #  temp_array_name as "", so that we can check while plotting between global and object
                            #  namespace data sets
                            temp_array_name = ""
                        # First, we handle the TypeError, and tell the user that the object name is invalid.
                        except TypeError:
                            clear_output()
                            # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                            print("\033[1mThe object name given is undefined. Make sure you've created an instance of the desired object!\033[0m")
                            continue
                        # Next, we handle the KeyError, and tell the user that the object has no array with the given name.
                        except KeyError:
                            clear_output()
                            # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                            print("\033[1mThe object name given is valid, but doesn't contain the array you've requested!\033[0m")
                            continue
                        # Finally, we handle the IndexError, and tell the user that the array they want exists but isn't
                        #  the correct size for the column requested
                        except IndexError:
                            clear_output()
                            # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                            print("\033[1mThe object and array names are both valid, but the array doesn't have the correct number of columns!\033[0m")
                            continue
                    # In this case, the variable requested by the user exists, but is not an array.
                    except TypeError:
                        clear_output()
                        # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                        print("\033[1mYou've entered the name of a variable that exists, but is not an array! Remember that lists will not work here, you must use an array that contains the time steps in the first column and the data in another column\033[0m")
                        continue
                    # In this case, we know that the array defined by the user is in the global namespace, but doesn't
                    #  have the correct number of columns.
                    except IndexError:
                        clear_output()
                        # Print the error message in bold (hence the strange escape codes at the start and end of this string)
                        print("\033[1mThe array you've requested exists, but it doesn't have the correct number of columns!\033[0m")
                        continue


                    # We now break out of the while True loop. This point will be reached if we either don't go through
                    #  the except KeyError statement (the array name given is in the global namespace),
                    #  or if we do go through it and don't raise any further errors (the array name given is in an object's namespace).
                    clear_output()
                    break

                # Now that we have a temporary variable containing an array name that we know is defined in the global
                #  namespace, let's append it to the list of array names containing the real-world data to graph.
                # The list will also contain "" for those graphs with no real-world data (if any), this way we can
                #  just check later whether it has a legitimate array name in each index, and if so, we graph it on the
                #  corresponding graph index
                self.real_data_arrays.append(temp_array_name)
                self.real_data_objects.append(temp_data_obj)
                self.real_data_objectArrays.append(temp_data_array)
                self.real_data_col_numbers.append(temp_column_number)


    # This is a local decorator to check whether any of the optional arguments to make_graphs have been passed
    #  explicitly by the user. Basically, it checks the arguments passed to the function it's wrapped around and then
    #  passes on the ones explicitly defined by the user as an argument to the wrapped function.
    # Look up documentation on decorator functions in Python for more information on how this works.
    def __explicit_checker(f):
        varnames = inspect.getfullargspec(f)[0]
        @wraps(f)
        def wrapper(self, *a, **kw):
            kw['explicit_params'] = set(list(varnames[:len(a)]) + list(kw.keys()))
            return f(self, *a, **kw)
        return wrapper

    # Decorate the function with the __explicit_checker function, so that whenever this function is called, we are actually
    #  calling __explicit_checker first.
    @__explicit_checker
    def make_graphs(self, figure_size = (20,20), std_dev=False, xaxis_labels = [], yaxis_labels = [], sims_line_labels=["", "", "", ""], sims_line_colors=["blue", "blue", "blue", "blue"], std_labels=["Mean +- Standard Deviation", "Mean +- Standard Deviation", "Mean +- Standard Deviation", "Mean +- Standard Deviation"], real_data_labels = ["", "", "", ""], real_data_colors = ["orange", "orange", "orange", "orange"], legend_locs=["lower right", "lower right", "lower right", "lower right"], graph_titles=[], savefile = "", explicit_params=None):
        """
        This method is what the user will call to actually create and display/save the graphs. It can take optional
        arguments that allow for making the graphs more presentable (things like axis labels and line styles/colors).

        The optional arguments are all passed in the form of lists with one entry per graph being created. So for graphs
        of CRH, ACTH and cortisol, if you wanted to pass titles for the graphs reflecting that fact, you'd use
            graph_titles = ["CRH Concentration", "ACTH Concentration", "Cortisol Concentration"]

        The full list of optional arguments to the function is as follows:
            sims_line_labels (to define labels for the simulated lines on each graph)
            sims_line_colors (to change the color of the simulated lines on each graph)
            std_labels (to define labels for the +- standard deviation shaded area)
            real_data_labels (to define labels for the real-world data on each graph, include list entries for
                each graph, with or without real-world data plotted)
            real_data_colors (to define line colors for the real-world data on each graph, again only include list entries
                for graphs with real-world data plotted)
            legend_locs (to change the location that the legend is displayed on each graph, if labels are added for the lines.
                Options for location are: "upper left", "upper right", "lower left", and "lower right")
            graph_titles (to add titles above the graphs)
            xaxis_labels (to change the labels on the x-axis)
            yaxis_labels (to change the labels on the y-axis)
            figure_size (to change the width and height of the figure, must be a tuple)
        """
        # Define the list of arrays we will fill with the organized data we want to plot. For no optimization runs,
        #  we will have num_iters = 0, so just replace it with 1 in this case so we have the properly sized array
        if self.num_iters > 0:
            self.sims = [np.zeros((len(self.globals_dict[self.array_name][:,0]), self.num_iters))]*self.num_graphs
        else:
            self.sims = [np.zeros((len(self.globals_dict[self.array_name][:,0]), 1))]*self.num_graphs

        # If we are plotting a parameter optimization run with more than one iteration, we need to gather the data in
        #  an organized fashion
        if self.num_iters > 1:
            # Loop through the variables the user has requested we graph
            for index, entry in enumerate(self.var_indices):
                # This may look a bit complex, but what we are doing is gathering the columns for each variable together.
                #  Before this, each variable is in columns like 1, 5, 9, etc. and we want to fix that so we can plot more easily
                #  So this statement loops through each column from the variable's first column (entry + 1) until the last column
                #  in the array, going by the number of columns per iteration each time (total columns/number of iterations)
                self.sims[index] = [self.globals_dict[self.array_name][:,i] for i in range(entry+1,len(self.globals_dict[self.array_name][0,:]),len(self.globals_dict[self.array_name][0,:]/self.num_iters))]
        # If we are plotting a parameter optimization run with a single iteration, or a simulation without optimization,
        #  put the data into self.sims, also
        else:
            for index, entry in enumerate(self.var_indices):
                self.sims[index] = self.globals_dict[self.array_name][:,entry+1]


        # Now that we have all of the data we want to graph organized and ready, we can define the figure and set the
        #  number of subplots in it to the number of variables the user has requested we graph
        fig, axes = plt.subplots(nrows = self.num_graphs, figsize = figure_size)

        # If the user is making multiple graphs, we do the following (where we loop through the subplots in the
        #  figure)
        if self.num_graphs > 1:
            # Next, we loop through the number of graphs requested by the user, and create them.
            for index, entry in enumerate(self.sims):
                # If we have multiple iterations, we will need to graph the mean (and optionally +- standard deviation),
                #  so check the number of iterations
                if self.num_iters > 1:
                    # Here we plot the mean over all iterations
                    axes[index].plot(self.globals_dict[self.array_name][:,0], np.mean(entry, axis = 0), label = sims_line_labels[index], color = sims_line_colors[index])

                    # Check if the user wants to plot the mean +- standard deviation, if so add it to the graph
                    if std_dev:
                        axes[index].fill_between(self.globals_dict[self.array_name][:,0], np.mean(entry, axis = 0) - np.std(entry, axis = 0), np.mean(entry, axis = 0) + np.std(entry, axis = 0), alpha = 0.3, label = std_labels[index])

                # If we have 0 or 1 iteration of parameter optimization, we can simply plot the data without taking mean
                #  or worrying about standard deviation
                else:
                    axes[index].plot(self.globals_dict[self.array_name][:,0], entry, label = sims_line_labels[index], color = sims_line_colors[index])

                # Now we check if the user has requested to plot any real-world data contained in the global namespace
                #  on this subplot, and graph it if so
                if self.real_data_arrays[index]:
                    axes[index].plot(self.globals_dict[self.real_data_arrays[index]][:,0], self.globals_dict[self.real_data_arrays[index]][:,self.real_data_col_numbers[index]], label = real_data_labels[index], color = real_data_colors[index])
                # If the user hasn't requested real-world data contained in the global namespace, we check if they have
                #  requested real-world data contained in an object's namespace. If so, we graph it.
                elif self.real_data_objects[index]:
                    axes[index].plot(vars(self.globals_dict[self.real_data_objects[index]])[self.real_data_objectArrays[index]][:,0], vars(self.globals_dict[self.real_data_objects[index]])[self.real_data_objectArrays[index]][:,self.real_data_col_numbers[index]], label = real_data_labels[index], color = real_data_colors[index])

                # Check if the user specified labels for the lines being graphed, if so we turn on the legend
                if 'sims_line_labels' in explicit_params:
                    try:
                        axes[index].legend(loc = legend_locs[index], shadow = True, fancybox = True)
                    except TypeError:
                        print("\033[1mMake sure you have passed a list with one entry per variable graphed for legend_locs\033[0m")
                # Check if the user specified x-axis labels, if so we turn them on
                if 'xaxis_labels' in explicit_params:
                    try:
                        axes[index].set(xlabel = xaxis_labels[index])
                    except TypeError:
                        print("\033[1mMake sure you have passed a list with one entry per variable graphed for xaxis_labels\033[0m")
                # Check if the user specified y-axis labels, if so we turn them on
                if 'yaxis_labels' in explicit_params:
                    try:
                        axes[index].set(ylabel = yaxis_labels[index])
                    except TypeError:
                        print("\033[1mMake sure you have passed a list with one entry per variable graphed for yaxis_labels\033[0m")
                # Check if the user specified titles for the graphs, if so we turn them on
                if 'graph_titles' in explicit_params:
                    try:
                        axes[index].set(title = graph_titles[index])
                    except TypeError:
                        print("\033[1mMake sure you have passed a list with one entry per variable graphed for graph_titles\033[0m")
        # If there's only one graph to make, do basically the same as above, just don't iterate through multiple
        #  variables and subplots
        else:
            if self.num_iters > 1:
                # Here we plot the mean over all iterations
                axes.plot(self.globals_dict[self.array_name][:,0], np.mean(self.sims[0], axis = 0), label = sims_line_labels[0], color = sims_line_colors[0])

                # Check if the user wants to plot the mean +- standard deviation, if so add it to the graph
                if std_dev:
                    axes.fill_between(self.globals_dict[self.array_name][:,0], np.mean(self.sims[0], axis = 0) - np.std(self.sims[0], axis = 0), np.mean(self.sims[0], axis = 0) + np.std(self.sims[0], axis = 0), alpha = 0.3, label = std_labels[0])
            # If we have 0 or 1 iteration of parameter optimization, we can simply plot the data without taking mean
            #  or worrying about standard deviation
            else:
                axes.plot(self.globals_dict[self.array_name][:,0], self.sims[0], label = sims_line_labels[0], color = sims_line_colors[0])

            # Now we check if the user has requested to plot any real-world data contained in the global namespace,
            #  and graph it if so
            if self.real_data_arrays[0]:
                axes.plot(self.globals_dict[self.real_data_arrays[0]][:,0], self.globals_dict[self.real_data_arrays[0]][:,self.real_data_col_numbers[0]], label = real_data_labels[0], color = real_data_colors[0])
            # If the user hasn't requested real-world data contained in the global namespace, we check if they have
            #  requested real-world data contained in an object's namespace. If so, we graph it.
            elif self.real_data_objects[0]:
                axes.plot(vars(self.globals_dict[self.real_data_objects[0]])[self.real_data_objectArrays[0]][:,0], vars(self.globals_dict[self.real_data_objects[0]])[self.real_data_objectArrays[0]][:,self.real_data_col_numbers[0]], label = real_data_labels[0], color = real_data_colors[0])

            # Check if the user specified labels for the lines being graphed, if so we turn on the legend
            if 'sims_line_labels' in explicit_params:
                try:
                    axes.legend(loc = legend_locs[0], shadow = True, fancybox = True)
                except TypeError:
                    print("\033[1mDespite only graphing a single variable, you still need to pass legend_locs as a list with one entry, sorry!\033[0m")
            # Check if the user specified x-axis labels, if so we turn them on
            if 'xaxis_labels' in explicit_params:
                try:
                    axes.set(xlabel = xaxis_labels[0])
                except TypeError:
                    print("\033[1mDespite only graphing a single variable, you still need to pass xaxis_labels as a list with one entry, sorry!\033[0m")
            # Check if the user specified y-axis labels, if so we turn them on
            if 'yaxis_labels' in explicit_params:
                try:
                    axes.set(ylabel = yaxis_labels[0])
                except TypeError:
                    print("\033[1mDespite only graphing a single variable, you still need to pass yaxis_labels as a list with one entry, sorry!\033[0m")
            # Check if the user specified titles for the graphs, if so we turn them on
            if 'graph_titles' in explicit_params:
                try:
                    axes.set(title = graph_titles[0])
                except TypeError:
                    print("\033[1mDespite only graphing a single variable, you still need to pass graph_titles as a list with one entry, sorry!\033[0m")

        # If the user has defined a filename, save the generated graphs with it
        if 'savefile' in explicit_params:
            plt.savefig(savefile, dpi=300)





# File Name: optimize.py
# Description: Contains the function to run parameter optimization algorithms
#  based on user inputs and return the cost function value, as well as the cost
#  function definitions that can be used.
# Author: Christopher Parker
# Created: Thu Aug 25, 2022 | 03:19P EDT
# Last Modified: Mon Sep 12, 2022 | 02:11P EDT

# Importing via __init__.py now
#import scipy.optimize as sco
#from scipy.interpolate import interp1d
#import time
#import numpy as np

# This function contains the main loop used for running iterations of 
#  parameter optimization algorithms.
# Non-optional arguments are: the cost function to use, the model function to use,
#  the data sets to optimize against, the initial conditions and the bounds on
#  parameters to optimize.
# Users can also optionally specify the number of iterations of the parameter
#  optimization algorithm to run, the algorithm to use (options include
#  differential_evolution, shgo, basinhopping and dual_annealing), the maximum
#  number of iterations of the algorithm to run INSIDE EACH FULL ITERATION, the
#  population size for the algorithm, the indices of the initial condition array
#  that are being optimized and the initial parameters to use (only for
#  basinhopping algorithm).
def run(cost_fun, model, data_to_match, y0, bounds, num_iter=5, \
        algorithm="differential_evolution", \
        maxiter=999, popsize=1, ICopt_indices = [], x0 = []):

    # Check the system clock time at the start of the loop so that we can keep
    #  track of how long the full run takes.
    start_time = time.time()

    # Run the parameter optimization algorithm num_iter times in a for loop
    for i in range(0,num_iter):

        # Let the user know which iteration the program is currently working on
        print(f"Optimization Run #{i+1}")

        # Check the algorithm the user wants to run, and call it
        if algorithm=="differential_evolution":
            res = sco.differential_evolution(cost_fun, bounds, \
                    maxiter = maxiter, disp = True, popsize = popsize)
        elif algorithm=="shgo":
            res = sco.shgo(cost_fun, bounds, options = {"f_min": 0.1, \
                            "maxiter": None, "minimize_every_iter": True, \
                            "local_iter": False, "disp": True}, iters = 3)
        elif algorithm=="basinhopping":
            res = sco.basinhopping(cost_fun, x0, niter = maxiter)
        elif algorithm=="dual_annealing":
            res = sco.dual_annealing(cost_fun, bounds)

        # Save the optimized initial conditions from the algorithm into the
        #  array y0 so that we can use it to re-run the model
        for index,entry in enumerate(ICopt_indices):
            y0[entry] = res.x[index]

        # We need to re-run the model with the optimized parameters to get the
        #  full variable concentration arrays, because we do not get this info
        #  from the parameter optimization algorithms themselves
        optimizedSimData = model(res.x[len(ICopt_indices):], y0)

        # Check if it's the first time through the loop
        if i == 0:
            # If it is, then initialize arrays to hold all of the concentration
            #  data and the optimized parameter values for all iterations
            simData_all = optimizedSimData
            opt_pars = np.hstack((res.fun, res.x[len(ICopt_indices):]))
        else:
            # If it's not the first time through the loop, then just add on to
            #  the arrays created in the first loop
            simData_all = np.hstack((simData_all, optimizedSimData))
            opt_pars_tmp = np.hstack((res.fun, res.x[len(ICopt_indices):]))
            opt_pars = np.vstack((opt_pars, opt_pars_tmp))

    # Check the clock time now that we've completed the loop
    end_time = time.time()
    # Report the total runtime
    print(f"Runtime: {end_time - start_time} seconds")

    # Return the optimized parameters from each iteration, and all of the
    #  concentration data from running the model with those optimized parameters
    return opt_pars, simData_all

# This function takes the time steps and data points for the real-world ACTH 
#  and cortisol concentration values you're optimizing against as well as the 
#  data from a simulation with the model and returns the cost for the parameter 
#  set used.
# There are optional arguments for ACTH_index and CORT_index in case the model
#  equations are ordered differently from the standard. That way the user can
#  pass the indices of ACTH and CORT as they are defined in y (or y0)
def SSE_cost(time_ACTH, data_ACTH, time_CORT, data_CORT, simData, 
             ACTH_index = 1, CORT_index = 2):
    # Compute the means of ACTH and CORT data arrays
    mean_ACTH = np.mean(data_ACTH)
    mean_CORT = np.mean(data_CORT)

    # Normalize the simData arrays by the mean values of data set to be matched.
    # We use ACTH_index and CORT_index plus 1 because the first column is the time
    # steps
    simNorm_ACTH = simData[:,ACTH_index+1]/mean_ACTH
    simNorm_CORT = simData[:,CORT_index+1]/mean_CORT

    # Normalize the data set to be matched, as well
    dataNorm_ACTH = data_ACTH/mean_ACTH
    dataNorm_CORT = data_CORT/mean_CORT

    # Now, we interpolate between the simulated data points, to ensure we have
    #  a simulated data point at the exact time of each real-world data point
    # We are currently doing a linear interpolation, although we could also
    #  choose to do cubic. however, I don't think it makes a huge amount of
    #  difference in this context
    spline_ACTH = interp1d(simData[:,0], simNorm_ACTH, kind = 'linear')
    spline_CORT = interp1d(simData[:,0], simNorm_CORT, kind = 'linear')

    # We use a try/except loop to ensure that we don't run into an error that
    #  causes the entire simulation to fail and have to start over
    try:
        # Compute the cost value for the current parameter set by finding the
        #  SSE between raw data and splines.
        acthSSE = np.sum((spline_ACTH(time_ACTH) - dataNorm_ACTH)**2)
        cortSSE = np.sum((spline_CORT(time_CORT) - dataNorm_CORT)**2)

        # We define cost as the average of the ACTH and CORT SSEs.
        cost = (acthSSE + cortSSE)/2

        return cost

    except ValueError:
        # In the case that we encounter a ValueError, this means that the ODE
        #  solver quit before computing all time steps up to t_end. The error
        #  arises when we attempt to plug in a time value after where the 
        #  solver quit to spline_ACTH or spline_CORT. So we catch the
        #  ValueError here, and let the user know that the ODE solver exited
        #  early.
        print("ODE solver did not make it through all data points.")


# This is similar to the function above, but does not take ACTH concentration
#  data as an input. So you'd use this one if your real-world data only has
#  cortisol concentrations
def SSE_cost_noACTH(time_CORT, data_CORT, simData, CORT_index = 2):
    # Compute the mean of the CORT data array
    mean_CORT = np.mean(data_CORT)

    # Normalize the simData array by the mean value of data set to be matched
    simNorm_CORT = simData[:,CORT_index+1]/mean_CORT

    # Normalize the data set to be matched, as well
    dataNorm_CORT = data_CORT/mean_CORT

    # Now, we interpolate between the simulated data points, to ensure we have
    #  a simulated data point at the exact time of each real-world data point
    # We are currently doing a linear interpolation, although we could also
    #  choose to do cubic. however, I don't think it makes a huge amount of
    #  difference in this context
    spline_CORT = interp1d(simData[:,0], simNorm_CORT, kind = 'linear')

    # We use a try/except loop to ensure that we don't run into an error that
    #  causes the entire simulation to fail and have to start over
    try:
        # As the data we are matching does not include ACTH values, we simply
        #  define cost as the SSE between raw data and splines of the CORT array
        cost = np.sum((spline_CORT(time_CORT) - dataNorm_CORT)**2)

        return cost

    except ValueError:
        # In the case that we encounter a ValueError, this means that the ODE
        #  solver quit before computing all time steps up to t_end. The error
        #  arises when we attempt to plug in a time value after where the 
        #  solver quit to spline_CORT. So we catch the ValueError here, and let 
        #  the user know that the ODE solver exited early.
        print("ODE solver did not make it through all data points.")

# We also need to define cost functions for using the minimum and maximum
#  distance between the simulation and real-world data.
#  
# Here, we use the maximum distance
def max_cost(time_ACTH, data_ACTH, time_CORT, data_CORT, simData,
             ACTH_index = 1, CORT_index = 2):
    # Compute the means of ACTH and CORT data arrays
    mean_ACTH = np.mean(data_ACTH)
    mean_CORT = np.mean(data_CORT)

    # Normalize the simData arrays by the mean values of data set to be matched
    simNorm_ACTH = simData[:,ACTH_index+1]/mean_ACTH
    simNorm_CORT = simData[:,CORT_index+1]/mean_CORT

    # Normalize the data set to be matched, as well
    dataNorm_ACTH = data_ACTH/mean_ACTH
    dataNorm_CORT = data_CORT/mean_CORT

    # Now, we interpolate between the simulated data points, to ensure we have
    #  a simulated data point at the exact time of each real-world data point
    # We are currently doing a linear interpolation, although we could also
    #  choose to do cubic. however, I don't think it makes a huge amount of
    #  difference in this context
    spline_ACTH = interp1d(simData[:,0], simNorm_ACTH, kind = 'linear')
    spline_CORT = interp1d(simData[:,0], simNorm_CORT, kind = 'linear')

    # We use a try/except loop to ensure that we don't run into an error that
    #  causes the entire simulation to fail and have to start over
    try:
        # Compute the cost value for the current parameter set by finding the
        #  maximum distance between the simulations and data for each of ACTH 
        #  and CORT
        acthMAX = np.amax((spline_ACTH(time_ACTH) - dataNorm_ACTH)**2)
        cortMAX = np.amax((spline_CORT(time_CORT) - dataNorm_CORT)**2)

        # We define cost as the maximum of the maximums that we just found above
        cost = np.amax([acthMAX, cortMAX])

        return cost

    except ValueError:
        # In the case that we encounter a ValueError, this means that the ODE
        #  solver quit before computing all time steps up to t_end. The error
        #  arises when we attempt to plug in a time value after where the 
        #  solver quit to spline_ACTH or spline_CORT. So we catch the
        #  ValueError here, and let the user know that the ODE solver exited
        #  early.
        print("ODE solver did not make it through all data points.")

# And here we compute the cost based on the maximum distance for data sets that
#  only contain cortisol data
def max_cost_noACTH(time_CORT, data_CORT, simData, CORT_index = 2):
    # Compute the mean of the CORT data array
    mean_CORT = np.mean(data_CORT)

    # Normalize the simData array by the mean value of data set to be matched
    simNorm_CORT = simData[:,CORT_index+1]/mean_CORT

    # Normalize the data set to be matched, as well
    dataNorm_CORT = data_CORT/mean_CORT

    # Now, we interpolate between the simulated data points, to ensure we have
    #  a simulated data point at the exact time of each real-world data point
    # We are currently doing a linear interpolation, although we could also
    #  choose to do cubic. however, I don't think it makes a huge amount of
    #  difference in this context
    spline_CORT = interp1d(simData[:,0], simNorm_CORT, kind = 'linear')

    # We use a try/except loop to ensure that we don't run into an error that
    #  causes the entire simulation to fail and have to start over
    try:
        # As the data we are matching does not include ACTH values, we simply
        #  define cost as the maximum distance between raw data and splines 
        #  of the CORT array
        cost = np.amax((spline_CORT(time_CORT) - dataNorm_CORT)**2)

        return cost

    except ValueError:
        # In the case that we encounter a ValueError, this means that the ODE
        #  solver quit before computing all time steps up to t_end. The error
        #  arises when we attempt to plug in a time value after where the 
        #  solver quit to spline_CORT. So we catch the ValueError here, and let 
        #  the user know that the ODE solver exited early.
        print("ODE solver did not make it through all data points.")



# File Name: __init__.py
# Description: Initialization for VeVaPy, allows importing of packages with
#  a single command. Also contains basic version and help information
# Author: Christopher Parker
# Created: Mon Jan 24, 2022 | 03:38P EST
# Last Modified: Mon Sep 12, 2022 | 02:12P EDT

__version__ = '0.1'

import matplotlib
matplotlib.use('inline')

import numpy as np
import scipy.integrate as sci
import scipy.optimize as sco
from scipy.interpolate import interp1d
import time
from IPython.display import clear_output
from functools import wraps
import matplotlib.pyplot as plt
import re

from VeVaPy import DEsolver
from VeVaPy import optimize
from VeVaPy.dataImport import data
from VeVaPy.visualize import Visualizer
